"""UNP-Profile."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field

import numpy as np
from matplotlib import pyplot as plt
from shapely.geometry import Polygon

from blueprints.structural_sections._polygon_builder import PolygonBuilder
from blueprints.structural_sections._profile import Profile
from blueprints.structural_sections.steel.profile_definitions.corrosion_utils import FULL_CORROSION_TOLERANCE, update_name_with_corrosion
from blueprints.structural_sections.steel.profile_definitions.plotters.general_steel_plotter import plot_shapes
from blueprints.type_alias import MM, PERCENTAGE
from blueprints.utils.math_helpers import slope_to_angle
from blueprints.validations import raise_if_negative


@dataclass(frozen=True, kw_only=True)
class UNPProfile(Profile):
    """Representation of U/C shaped profiles.

    For standard profiles, use the specific standard profile class like `UNP`. For example,
    ```python
    unp_profile = UNP.UNP200
    ```

    Attributes
    ----------
    top_flange_total_width : MM
        The total width of the top flange, measured halfway the total width of the element [mm].
    top_flange_thickness : MM
        The thickness of the top flange [mm].
    bottom_flange_total_width : MM
        The total width of the bottom flange, measured halfway the total width of the element [mm].
    bottom_flange_thickness : MM
        The thickness of the bottom flange [mm].
    total_height : MM
        The total height of the profile [mm].
    web_thickness : MM
        The thickness of the web [mm].
    top_root_fillet_radius : MM
        The radius of the curved corners of the top flange. Default is 0, meaning sharp corner.
    top_toe_radius : MM
        The radius of the outer corners of the top flange. Default is 0, meaning sharp corner.
    top_outer_corner_radius : MM
        The radius of the outer corners of the top flange. Default is 0, meaning sharp corner.
    bottom_root_fillet_radius : MM
        The radius of the curved corners of the bottom flange. Default is 0, meaning sharp corner.
    bottom_toe_radius : MM
        The radius of the outer corners of the bottom flange. Default is 0, meaning sharp corner.
    bottom_outer_corner_radius : MM
        The radius of the outer corners of the bottom flange. Default is 0, meaning sharp corner.
    top_slope : PERCENTAGE
        The slope of the top flange. Default is 0.
    bottom_slope : PERCENTAGE
        The slope of the bottom flange. Default is 0.
    name : str
        The name of the profile. Default is "UNP-Profile". If corrosion is applied, the name will include the corrosion value.
    plotter : Callable[[Profile], plt.Figure]
        The plotter function to visualize the profile (default: `plot_shapes`).
    """

    top_flange_total_width: MM
    """The total width of the top flange [mm]."""
    top_flange_thickness: MM
    """The thickness of the top flange [mm]."""
    bottom_flange_total_width: MM
    """The total width of the bottom flange [mm]."""
    bottom_flange_thickness: MM
    """The thickness of the bottom flange [mm]."""
    total_height: MM
    """The total height of the profile [mm]."""
    web_thickness: MM
    """The thickness of the web [mm]."""
    top_root_fillet_radius: MM = 0.0
    """The radius of the curved corners of the top flange. Default is 0, meaning sharp corner."""
    top_toe_radius: MM = 0.0
    """The radius of the outer corners of the top flange. Default is 0, meaning sharp corner."""
    top_outer_corner_radius: MM = 0.0
    """The radius of the outer corners of the top flange. Default is 0, meaning sharp corner."""
    bottom_root_fillet_radius: MM = 0.0
    """The radius of the curved corners of the bottom flange. Default is 0, meaning sharp corner."""
    bottom_toe_radius: MM = 0.0
    """The radius of the outer corners of the bottom flange. Default is 0, meaning sharp corner."""
    bottom_outer_corner_radius: MM = 0.0
    """The radius of the outer corners of the bottom flange. Default is 0, meaning sharp corner."""
    top_slope: PERCENTAGE = 0.0
    """The slope of the top flange. Default is 0."""
    bottom_slope: PERCENTAGE = 0.0
    """The slope of the bottom flange. Default is 0."""
    name: str = "UNP-Profile"
    """The name of the profile. Default is "UNP-Profile". If corrosion is applied, the name will include the corrosion value."""
    plotter: Callable[[Profile], plt.Figure] = plot_shapes
    """The plotter function to visualize the profile (default: `plot_shapes`)."""

    # Attributes set in __post_init__ (for type checkers)
    top_root_fillet_height: float = field(init=False, repr=False)
    top_root_fillet_width: float = field(init=False, repr=False)
    bottom_root_fillet_height: float = field(init=False, repr=False)
    bottom_root_fillet_width: float = field(init=False, repr=False)
    top_toe_radius_height: float = field(init=False, repr=False)
    top_toe_radius_width: float = field(init=False, repr=False)
    bottom_toe_radius_height: float = field(init=False, repr=False)
    bottom_toe_radius_width: float = field(init=False, repr=False)
    top_slope_width: float = field(init=False, repr=False)
    top_slope_height: float = field(init=False, repr=False)
    top_slope_length: float = field(init=False, repr=False)
    bottom_slope_width: float = field(init=False, repr=False)
    bottom_slope_height: float = field(init=False, repr=False)
    bottom_slope_length: float = field(init=False, repr=False)
    top_toe_total_height: float = field(init=False, repr=False)
    top_toe_flat_height: float = field(init=False, repr=False)
    bottom_toe_total_height: float = field(init=False, repr=False)
    bottom_toe_flat_height: float = field(init=False, repr=False)
    web_inner_height_top: float = field(init=False, repr=False)
    web_inner_height_bottom: float = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Post-initialization: calculate corner and slope parameters for the UNP profile."""
        object.__setattr__(self, "top_root_fillet_height", np.cos(np.deg2rad(slope_to_angle(self.top_slope))) * self.top_root_fillet_radius)
        object.__setattr__(self, "top_root_fillet_width", (1 - np.sin(np.deg2rad(slope_to_angle(self.top_slope)))) * self.top_root_fillet_radius)

        object.__setattr__(self, "bottom_root_fillet_height", np.cos(np.deg2rad(slope_to_angle(self.bottom_slope))) * self.bottom_root_fillet_radius)
        object.__setattr__(
            self, "bottom_root_fillet_width", (1 - np.sin(np.deg2rad(slope_to_angle(self.bottom_slope)))) * self.bottom_root_fillet_radius
        )

        object.__setattr__(self, "top_toe_radius_height", np.cos(np.deg2rad(slope_to_angle(self.top_slope))) * self.top_toe_radius)
        object.__setattr__(self, "top_toe_radius_width", (1 - np.sin(np.deg2rad(slope_to_angle(self.top_slope)))) * self.top_toe_radius)

        object.__setattr__(self, "bottom_toe_radius_height", np.cos(np.deg2rad(slope_to_angle(self.bottom_slope))) * self.bottom_toe_radius)
        object.__setattr__(self, "bottom_toe_radius_width", (1 - np.sin(np.deg2rad(slope_to_angle(self.bottom_slope)))) * self.bottom_toe_radius)

        object.__setattr__(
            self, "top_slope_width", self.top_flange_total_width - self.web_thickness - self.top_root_fillet_width - self.top_toe_radius_width
        )
        object.__setattr__(self, "top_slope_height", np.tan(np.deg2rad(slope_to_angle(self.top_slope))) * self.top_slope_width)
        object.__setattr__(self, "top_slope_length", np.sqrt(self.top_slope_width**2 + self.top_slope_height**2))

        object.__setattr__(
            self,
            "bottom_slope_width",
            self.bottom_flange_total_width - self.web_thickness - self.bottom_root_fillet_width - self.bottom_toe_radius_width,
        )
        object.__setattr__(self, "bottom_slope_height", np.tan(np.deg2rad(slope_to_angle(self.bottom_slope))) * self.bottom_slope_width)
        object.__setattr__(self, "bottom_slope_length", np.sqrt(self.bottom_slope_width**2 + self.bottom_slope_height**2))

        object.__setattr__(
            self,
            "top_toe_total_height",
            self.top_flange_thickness - (self.top_flange_total_width / 2 - self.top_toe_radius_width) * self.top_slope / 100,
        )
        object.__setattr__(self, "top_toe_flat_height", self.top_toe_total_height - self.top_toe_radius_height)

        object.__setattr__(
            self,
            "bottom_toe_total_height",
            self.bottom_flange_thickness - (self.bottom_flange_total_width / 2 - self.bottom_toe_radius_width) * self.bottom_slope / 100,
        )
        object.__setattr__(self, "bottom_toe_flat_height", self.bottom_toe_total_height - self.bottom_toe_radius_height)

        object.__setattr__(
            self, "web_inner_height_top", self.total_height / 2 - self.top_toe_total_height - self.top_slope_height - self.top_root_fillet_height
        )
        object.__setattr__(
            self,
            "web_inner_height_bottom",
            self.total_height / 2 - self.bottom_toe_total_height - self.bottom_slope_height - self.bottom_root_fillet_height,
        )

        raise_if_negative(
            top_slope=self.top_slope,
            bottom_slope=self.bottom_slope,
            top_root_fillet_height=self.top_root_fillet_height,
            top_root_fillet_width=self.top_root_fillet_width,
            bottom_root_fillet_height=self.bottom_root_fillet_height,
            bottom_root_fillet_width=self.bottom_root_fillet_width,
            top_toe_radius_height=self.top_toe_radius_height,
            top_toe_radius_width=self.top_toe_radius_width,
            bottom_toe_radius_height=self.bottom_toe_radius_height,
            bottom_toe_radius_width=self.bottom_toe_radius_width,
            top_slope_width=self.top_slope_width,
            top_slope_height=self.top_slope_height,
            top_slope_length=self.top_slope_length,
            bottom_slope_width=self.bottom_slope_width,
            bottom_slope_height=self.bottom_slope_height,
            bottom_slope_length=self.bottom_slope_length,
            top_toe_total_height=self.top_toe_total_height,
            top_toe_flat_height=self.top_toe_flat_height,
            bottom_toe_total_height=self.bottom_toe_total_height,
            bottom_toe_flat_height=self.bottom_toe_flat_height,
            web_inner_height_top=self.web_inner_height_top,
            web_inner_height_bottom=self.web_inner_height_bottom,
        )

    @property
    def max_thickness(self) -> MM:
        """Maximum element thickness of the profile [mm]."""
        return max(self.top_flange_thickness, self.bottom_flange_thickness, self.web_thickness)

    @property
    def _polygon(self) -> Polygon:
        """Return the polygon of the UNP profile without the offset and rotation applied."""
        return (
            PolygonBuilder(starting_point=(0, 0))
            # Starting halfway along the web, going up
            .append_line(length=self.total_height / 2 - self.top_outer_corner_radius, angle=90)
            # Top flange
            .append_arc(sweep=-90, angle=90, radius=self.top_outer_corner_radius)
            .append_line(length=self.top_flange_total_width - self.top_outer_corner_radius, angle=0)
            .append_line(length=self.top_toe_flat_height, angle=270)
            .append_arc(sweep=-90 + slope_to_angle(self.top_slope), angle=270, radius=self.top_toe_radius)
            .append_line(length=self.top_slope_length, angle=180 + slope_to_angle(self.top_slope))
            .append_arc(sweep=90 - slope_to_angle(self.top_slope), angle=180 + slope_to_angle(self.top_slope), radius=self.top_root_fillet_radius)
            # Going down the web
            .append_line(length=self.web_inner_height_top, angle=270)
            .append_line(length=self.web_inner_height_bottom, angle=270)
            # Bottom flange
            .append_arc(sweep=90 - slope_to_angle(self.bottom_slope), angle=270, radius=self.bottom_root_fillet_radius)
            .append_line(length=self.bottom_slope_length, angle=-slope_to_angle(self.bottom_slope))
            .append_arc(sweep=-90 + slope_to_angle(self.bottom_slope), angle=-slope_to_angle(self.bottom_slope), radius=self.bottom_toe_radius)
            .append_line(length=self.bottom_toe_flat_height, angle=270)
            .append_line(length=self.bottom_flange_total_width - self.bottom_outer_corner_radius, angle=180)
            .append_arc(sweep=-90, angle=180, radius=self.bottom_outer_corner_radius)
            # Closing the profile
            .append_line(length=self.total_height / 2 - self.bottom_outer_corner_radius, angle=90)
            .generate_polygon()
        )

    def with_corrosion(self, corrosion: MM = 0) -> UNPProfile:
        """Apply corrosion to the UNP-profile and return a new UNP-profile instance.

        The name attribute of the new instance will be updated to reflect the total corrosion applied
        including any previous corrosion indicated in the original name.

        Parameters
        ----------
        corrosion : MM, optional
            Corrosion per side (default is 0).

        Returns
        -------
        UNPProfile
            A new UNP-profile instance with the applied corrosion.

        Raises
        ------
        ValueError
            If the profile has fully corroded.
        """
        raise_if_negative(corrosion=corrosion)

        if corrosion == 0:
            return self

        # Use a buffer dict to store updated dimensions
        buffer = {
            "top_flange_total_width": self.top_flange_total_width - 2 * corrosion,
            "top_flange_thickness": self.top_flange_thickness - 2 * corrosion,
            "bottom_flange_total_width": self.bottom_flange_total_width - 2 * corrosion,
            "bottom_flange_thickness": self.bottom_flange_thickness - 2 * corrosion,
            "total_height": self.total_height - 2 * corrosion,
            "web_thickness": self.web_thickness - 2 * corrosion,
            "top_root_fillet_radius": self.top_root_fillet_radius + corrosion,
            "top_toe_radius": max(0, self.top_toe_radius - corrosion),
            "top_outer_corner_radius": max(0, self.top_outer_corner_radius - corrosion),
            "bottom_root_fillet_radius": self.bottom_root_fillet_radius + corrosion,
            "bottom_toe_radius": max(0, self.bottom_toe_radius - corrosion),
            "bottom_outer_corner_radius": max(0, self.bottom_outer_corner_radius - corrosion),
        }

        if any(
            buffer[thickness] < FULL_CORROSION_TOLERANCE
            for thickness in (
                "top_flange_thickness",
                "bottom_flange_thickness",
                "web_thickness",
            )
        ):
            raise ValueError("The profile has fully corroded.")

        name = update_name_with_corrosion(self.name, corrosion=corrosion)

        return UNPProfile(
            top_flange_total_width=buffer["top_flange_total_width"],
            top_flange_thickness=buffer["top_flange_thickness"],
            bottom_flange_total_width=buffer["bottom_flange_total_width"],
            bottom_flange_thickness=buffer["bottom_flange_thickness"],
            total_height=buffer["total_height"],
            web_thickness=buffer["web_thickness"],
            top_root_fillet_radius=buffer["top_root_fillet_radius"],
            top_toe_radius=buffer["top_toe_radius"],
            top_outer_corner_radius=buffer["top_outer_corner_radius"],
            bottom_root_fillet_radius=buffer["bottom_root_fillet_radius"],
            bottom_toe_radius=buffer["bottom_toe_radius"],
            bottom_outer_corner_radius=buffer["bottom_outer_corner_radius"],
            top_slope=self.top_slope,
            bottom_slope=self.bottom_slope,
            name=name,
            plotter=self.plotter,
        )



"""Testing Formula E.1 from EN 1995-1-1:2023."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1995_1_1_2023.appendix_e.formula_e_1 import FormEDot1EffBendingStiffness
from blueprints.validations import LessOrEqualToZeroError


class TestFormEDot1EffBendingStiffnessLi:
    """Validation for formula E.1 from EN 1995-1-1:2023."""

    @pytest.mark.parametrize(
        ("e_i", "i_i", "gamma_i", "a_i", "alpha_i", "expected_result"),
        [
            ([1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], 40),
            ([1, 1, 200], [1, 1, 400], [1, 1, 0.2], [1, 1, 0.5], [1, 1, 0.25], 80005.250),
            ([1, 1], [1, 1], [1, 1], [1, 1], [1, 1], 4),
            ([1, 1], [1, 1], [1, 1], [1, 1], [1, 0], 3),
        ],
        ids=["test-3arg", "test-3argb", "test-2arg", "test-2argb"],
    )
    def test_evaluation(
        self, e_i: list[float], i_i: list[float], gamma_i: list[float], a_i: list[float], alpha_i: list[float], expected_result: float
    ) -> None:
        """Test the evaluation of the result."""
        form_e_1 = FormEDot1EffBendingStiffness(e_i=e_i, i_i=i_i, gamma_i=gamma_i, a_i=a_i, alpha_i=alpha_i)
        assert form_e_1 == pytest.approx(expected=expected_result, abs=1e-3)

    @pytest.mark.parametrize(
        ("e_i", "i_i", "gamma_i", "a_i", "alpha_i", "expectation"),
        [
            ([1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], does_not_raise()),
            ([1, 1], [1, 1], [1, 1], [1, 1], [1, 0], does_not_raise()),
            ([1, 1], [1, 1], [1, -1], [1, 1], [1, 0], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1], [1, 1], [1, 1], [-1, 1], [1, 0], pytest.raises(LessOrEqualToZeroError)),
            ([-1, 1], [1, 1], [1, 1], [1, 1], [1, 0], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1], [0, 1], [1, 1], [1, 1], [1, 0], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1, 0], [1, 1], [1, 1], [1, 1], [1, 0], pytest.raises(ValueError, match="All input lists must have the same length")),
            (
                [1, 1, 0, 2],
                [1, 1, 2, 3],
                [1, 1, 2, 1],
                [1, 1, 1, 1],
                [1, 0, 1, 1],
                pytest.raises(ValueError, match="The length of the lists/layers must be either 2 or 3"),
            ),
        ],
        ids=["integer number", "decimal number", "negative_gamma_i", "negative_A_i", "negative_E_i", "zero_I_i", "unequal_lists", "long lists"],
    )
    def test_raise_error_incorrect_args(
        self, e_i: list[float], i_i: list[float], gamma_i: list[float], a_i: list[float], alpha_i: list[float], expectation: AbstractContextManager
    ) -> None:
        """Test the evaluation of incorrect arguments."""
        with expectation:
            assert FormEDot1EffBendingStiffness(e_i=e_i, i_i=i_i, gamma_i=gamma_i, a_i=a_i, alpha_i=alpha_i) is not None

    @pytest.mark.parametrize(
        ("e_i", "i_i", "gamma_i", "a_i", "alpha_i", "rep_short", "rep_long"),
        [
            (
                [1, 1],
                [1, 1],
                [1, 1],
                [1, 1],
                [1, 1],
                r"(EI)_{ef} = 4.00",
                r"(EI)_{ef} = (E_1 I_1 + \gamma_1 E_1 A_1 \alpha_1^2) + (E_2 I_2 + \gamma_2 E_2 A_2 \alpha_2^2) = "
                r"(1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00^2) + "
                r"(1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00^2) = 4.00",
            ),
            (
                [1, 1, 200],
                [1, 1, 400],
                [1, 1, 0.2],
                [1, 1, 0.5],
                [1, 1, 0.25],
                r"(EI)_{ef} = 80005.25",
                r"(EI)_{ef} = (E_1 I_1 + \gamma_1 E_1 A_1 \alpha_1^2) + (E_2 I_2 + \gamma_2 E_2 A_2 \alpha_2^2) + "
                r"(E_3 I_3 + \gamma_3 E_3 A_3 \alpha_3^2) = "
                r"(1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00^2) + "
                r"(1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00^2) + "
                r"(200.00 \cdot 400.00 + 0.20 \cdot 200.00 \cdot 0.50 \cdot 0.25^2) = 80005.25",
            ),
        ],
        ids=["length_2", "length_3"],
    )
    def test_latex(
        self, e_i: list[float], i_i: list[float], gamma_i: list[float], a_i: list[float], alpha_i: list[float], rep_short: str, rep_long: str
    ) -> None:
        """Test the latex representation of the formula."""
        form_e_1_a_latex = FormEDot1EffBendingStiffness(e_i=e_i, i_i=i_i, gamma_i=gamma_i, a_i=a_i, alpha_i=alpha_i).latex()
        assert form_e_1_a_latex.complete == rep_long
        assert form_e_1_a_latex.short == rep_short



"""Helper methods for building standardized report sections."""

from collections.abc import Sequence
from typing import Any, Optional

from blueprints.checks.check_protocol import CheckProtocol
from blueprints.checks.check_result import CheckResult
from blueprints.saf.results.result_internal_force_1d import ResultInternalForce1D
from blueprints.structural_sections.steel.steel_cross_section import SteelCrossSection
from blueprints.utils.report import Report


def add_applied_documents(report: Report, docs: Sequence[Any]) -> None:
    """
    Add applied code documents section to the report.

    report : Report
        The report object to which the calculation steps will be added.
    docs : Sequence[Any]
        The list of applied code documents.
    """
    report.add_heading("Applied code documents")
    report.add_paragraph("The following documents were applied in this check:")
    report.add_list(docs)


def add_applied_forces(report: Report, forces: ResultInternalForce1D, n: int = 2) -> None:
    """
    Add applied forces section to the report.

    report : Report
        The report object to which the calculation steps will be added.
    forces : ResultInternalForce1D
        The internal forces object.
    n : int, optional
        Number of decimals (default is 2).
    """
    force_labels = {
        "n": "Normal force $N$ [kN]",
        "vz": "Shear force $V_z$ [kN]",
        "vy": "Shear force $V_y$ [kN]",
        "my": "Bending moment $M_y$ [kNm]",
        "mz": "Bending moment $M_z$ [kNm]",
        "mx": "Torsion $T$ [kNm]",
    }

    report.add_heading("Applied internal forces")
    report.add_paragraph("The following internal forces were applied in this check:")
    rows = []
    for force_attr in ["n", "vy", "vz", "mx", "my", "mz"]:
        value = getattr(forces, force_attr, None)
        if value is not None:
            label = force_labels[force_attr] if force_labels and force_attr in force_labels else force_attr.upper()
            rows.append([label, f"{value:.{n}f}"])
    report.add_table(headers=["Internal Force", "Value"], rows=rows)


def add_material_steel_info(report: Report, steel_cross_section: SteelCrossSection, n: int = 2) -> None:
    """
    Add material and steel info to the report.

    report : Report
        The report object to which the calculation steps will be added.
    steel_cross_section : SteelCrossSection
        The steel cross section object.
    n : int, optional
        Number of decimals (default is 2).
    """
    report.add_heading("Applied material and profile")
    report.add_paragraph("The following material properties were used in this check:")
    mat = steel_cross_section.material
    report.add_table(
        headers=["Property", "Value"],
        rows=[
            ["Material", str(getattr(mat, "name", mat))],
            ["Yield Strength $f_y$", f"{getattr(steel_cross_section, 'yield_strength', 0):.{n}f} MPa"],
            ["Ultimate Strength $f_u$", f"{getattr(steel_cross_section, 'ultimate_strength', 0):.{n}f} MPa"],
            ["Elastic Modulus $E$", f"{getattr(mat, 'e_modulus', 0):.{n}f} MPa"],
        ],
    )


def add_section_properties(
    report: Report,
    section_properties: object,
    profile: object = None,
    n: int = 2,
    properties: Sequence[str] = ("area",),
) -> None:
    """
    Add section properties to the report.

    report : Report
        The report object to which the calculation steps will be added.
    section_properties : object
        The section properties object.
    profile : object, optional
        Optional profile object for name.
    n : int, optional
        Number of decimals (default is 2).
    properties : Sequence[str], optional
        List of property names to include (default is ("area",)).
    """
    # Mini dictionary of common property labels used in section reports
    # please note: sectionproperties uses x-y coordinates, while Blueprints y-z as in structural analysis
    property_labels = {
        # mm^2
        "area": "Area $A$ in $mm^{2}$",
        "a_sx": "Shear area $A_{sy}$ in $mm^{2}$",
        "a_sy": "Shear area $A_{sz}$ in $mm^{2}$",
        # mm
        "perimeter": "Perimeter $P$ in $mm$",
        # kg/m
        "mass": "Mass $W$ in $kg/m$",
        # mm^3
        "zxx_plus": "Section modulus $W_{y,el+}$ in $mm^{3}$",
        "zxx_minus": "Section modulus $W_{y,el-}$ in $mm^{3}$",
        "zyy_plus": "Section modulus $W_{z,el+}$ in $mm^{3}$",
        "zyy_minus": "Section modulus $W_{z,el-}$ in $mm^{3}$",
        "sxx": "Plastic section modulus $W_{y,pl}$ in $mm^{3}$",
        "syy": "Plastic section modulus $W_{z,pl}$ in $mm^{3}$",
        # mm^4
        "ixx_c": "Second moment of area $I_{y}$ in $mm^{4}$",
        "iyy_c": "Second moment of area $I_{z}$ in $mm^{4}$",
        "j": "Torsion constant $I_{T}$ in $mm^{4}$",
    }

    report.add_paragraph("The following section properties were used in this check:")
    rows = []
    if profile is not None:
        rows.append(["Profile", str(getattr(profile, "name", profile))])
    for prop in properties:
        label = property_labels[prop] if property_labels and prop in property_labels else prop.capitalize()
        value = getattr(section_properties, prop, None)
        if value is None:
            rows.append([label, "Not calculated"])
        else:
            value_str = f"{value:.{n}f}" if isinstance(value, float) else str(value)
            rows.append([label, value_str])

    report.add_table(headers=["Property", "Value"], rows=rows)


def add_unity_check_summary(report: Report, calculations: dict[str, Optional["CheckProtocol"] | CheckResult], n: int = 2) -> None:
    """
    Add a summary table of unity checks to the report.

    report : Report
        The report object to which the calculation steps will be added.
    calculations : dict[str, "CheckProtocol" | CheckResult]
        Iterable of (check_name, check_instance) pairs where check_instance can be either
        a CheckProtocol object or a CheckResult object (e.g., dict.items() or list of tuples).
    n : int, optional
        Number of decimals (default is 2).
    """
    report.add_heading("Utilization summary")
    rows = []
    overall_ok = True
    for check_name, check in calculations.items():
        if check is None:
            continue
        if isinstance(check, CheckResult):
            continue

        res = check.result()
        uc = getattr(res, "unity_check", None)
        utilization = f"{uc:.{n}f}" if uc is not None else "Not calculated"
        status = "OK" if getattr(res, "is_ok", False) else "NOT OK"
        if not getattr(res, "is_ok", False):
            overall_ok = False
        rows.append([check_name.capitalize(), utilization, status])
    if len(rows) == 0:
        report.add_paragraph("No checks were performed.")
        return
    report.add_table(headers=["Check", "Utilization", "Status"], rows=rows)
    report.add_paragraph(f"Overall result: {'OK' if overall_ok else 'NOT OK'}", bold=True)



"""Defines a general steel plotter for profiles and its characteristics."""

import matplotlib.pyplot as plt
from matplotlib import patches as mplpatches
from matplotlib.patches import Polygon as MplPolygon
from shapely.geometry import Point

from blueprints.structural_sections._profile import Profile

# Define color
STEEL_COLOR = (0.683, 0.0, 0.0)


def plot_shapes(
    profile: Profile,
    figsize: tuple[float, float] = (15.0, 8.0),
    title: str = "",
    font_size_title: float = 18.0,
    font_size_legend: float = 10.0,
    show: bool = False,
    include_moment_of_inertia: bool = False,
) -> plt.Figure:
    """
    Plot the given shapes.

    Parameters
    ----------
    profile : Profile
        The profile to plot.
    figsize : tuple[float, float], optional
        The size of the figure in inches. Default is (15.0, 8.0).
    title : str, optional
        The title of the plot. Default is "".
    font_size_title : float, optional
        The font size of the title. Default is 18.0.
    font_size_legend : float, optional
        The font size of the legend. Default is 10.0.
    show : bool, optional
        Whether to show the plot. Default is False.
    include_moment_of_inertia : bool, optional
        Whether to include the moment of inertia in the legend. Default is False.
    """
    fig, ax = plt.subplots(figsize=figsize)

    # Plot the exterior polygon
    coords = profile.polygon.exterior.coords
    patch = MplPolygon(xy=coords, lw=1, fill=True, facecolor=STEEL_COLOR, edgecolor=STEEL_COLOR)
    ax.add_patch(patch)

    # Plot the interior polygons (holes) if any
    for interior in profile.polygon.interiors:
        coords = interior.coords
        patch = MplPolygon(xy=coords, lw=0, fill=True, facecolor="white")
        ax.add_patch(patch)

    # Add dimension lines and centroid
    _add_dimension_lines(ax=ax, profile=profile, centroid=profile.centroid)
    ax.plot(profile.centroid.x, profile.centroid.y, "o", color="black")

    # Add legend text
    legend_text = f"""
    {profile.name}\n
    Area: {profile.area:.1f} mm²"""

    if include_moment_of_inertia:
        profile_section_properties = profile.section_properties(plastic=False, warping=False)
        legend_text += f"""
    Moment of inertia about x: {profile_section_properties.ixx_c:.0f} mm⁴
    Moment of inertia about y: {profile_section_properties.iyy_c:.0f} mm⁴
    """

    # Get the boundaries of the plot
    _, min_y, max_x, _ = profile.polygon.bounds
    offset = profile.profile_width / 20

    # Add the legend text to the plot
    ax.annotate(
        xy=(max_x + offset, min_y),
        text=legend_text,
        transform=ax.transAxes,
        fontsize=font_size_legend,
    )

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_title(title, fontsize=font_size_title)
    ax.grid(True)
    ax.axis("equal")
    ax.axis("off")
    ax.set_xlim(profile.polygon.bounds[0] - offset, profile.polygon.bounds[2] + offset)

    if show:
        plt.show()  # pragma: no cover

    assert fig is not None

    return fig


def _add_dimension_lines(ax: plt.Axes, profile: Profile, centroid: Point) -> None:
    """Adds dimension lines to show the outer dimensions of the geometry.

    Parameters
    ----------
    ax : plt.Axes
        The matplotlib axes to draw on.
    profile : tuple[Profile, ...]
        The profiles to plot.
    centroid : Point
        The centroid of the profile.
    """
    # Define the offset for the dimension lines
    offset_dimension_lines = max(profile.profile_height, profile.profile_width) / 20
    offset_text = offset_dimension_lines / 2

    # Calculate the bounds of all elements in the geometry
    min_x, min_y, max_x, max_y = profile.polygon.bounds

    # Calculate the width and height of the geometry relative to the centroid
    centroid_width = centroid.x - min_x
    centroid_height = centroid.y - min_y

    # Add the width dimension line (below the geometry)
    diameter_line_style = {
        "arrowstyle": mplpatches.ArrowStyle(stylename="<->", head_length=0.5, head_width=0.5),
    }

    # HORIZONTAL DIMENSION LINES (BELOW THE GEOMETRY)
    # Add the width dimension lines (below the geometry)
    ax.annotate(
        text="",
        xy=(min_x, min_y - offset_dimension_lines * 2),
        xytext=(max_x, min_y - offset_dimension_lines * 2),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        annotation_clip=False,
    )
    ax.text(
        s=f"b= {profile.profile_width:.1f} mm",
        x=(min_x + max_x) / 2,
        y=min_y - offset_dimension_lines * 2 + offset_text,
        verticalalignment="top",
        horizontalalignment="center",
        fontsize=10,
    )

    # Add the width dimension lines (below the geometry)
    ax.annotate(
        text="",
        xy=(min_x, min_y - offset_dimension_lines),
        xytext=(centroid.x, min_y - offset_dimension_lines),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        annotation_clip=False,
    )
    ax.text(
        s=f"{centroid_width:.1f} mm",
        x=(min_x + centroid.x) / 2,
        y=min_y - offset_dimension_lines + offset_text,
        verticalalignment="top",
        horizontalalignment="center",
        fontsize=10,
    )

    # HEIGHT DIMENSION LINES (ON THE LEFT SIDE OF THE GEOMETRY)
    # Add the distance from the bottom to the centroid (on the left side)
    ax.annotate(
        text="",
        xy=(min_x - offset_dimension_lines, min_y),
        xytext=(min_x - offset_dimension_lines, centroid.y),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        rotation=90,
        annotation_clip=False,
    )
    ax.text(
        s=f"{centroid_height:.1f} mm",
        x=min_x - offset_dimension_lines - offset_text,
        y=(min_y + centroid.y) / 2,
        verticalalignment="center",
        horizontalalignment="left",
        fontsize=10,
        rotation=90,
    )

    # # Add the height dimension line (on the left side of the geometry)
    ax.annotate(
        text="",
        xy=(min_x - offset_dimension_lines * 2, max_y),
        xytext=(min_x - offset_dimension_lines * 2, min_y),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        rotation=90,
        annotation_clip=False,
    )
    ax.text(
        s=f"h= {profile.profile_height:.1f} mm",
        x=(min_x - offset_dimension_lines * 2 - offset_text),
        y=(min_y + max_y) / 2,
        verticalalignment="center",
        horizontalalignment="left",
        fontsize=10,
        rotation=90,
    )



"""Resultant internal forces used for strength or stability checks.

Internal forces on line, beam, member. Result in member axis (not in principal axis).
"""

from dataclasses import dataclass
from enum import Enum

from blueprints.type_alias import KN, KNM, M


class ResultOn(str, Enum):
    """Enumeration for where the internal force result is located.

    Following SAF standard specification for ResultInternalForce1D.
    """

    ON_BEAM = "On beam"
    ON_RIB = "On rib"


class ResultFor(str, Enum):
    """Enumeration for the source type of the internal force result.

    Following SAF standard specification for ResultInternalForce1D.
    """

    LOAD_CASE = "Load case"
    LOAD_COMBINATION = "Load combination"


@dataclass(frozen=True)
class ResultInternalForce1D:
    """Internal forces on line, beam, member. Result in member axis (not in principal axis).

    Definition following https://www.saf.guide/en/stable/results/resultinternalforce1d.html.

    Coordinate System:

        z (vertical, usually strong axis)
            ↑
            |     x (longitudinal beam direction, into screen)
            |    ↗
            |   /
            |  /
            | /
            |/
      ←-----O
       y (horizontal/side, usually weak axis)

    Sign Conventions:
    - Normal force N_x: positive = tension, negative = compression.
    - Shear force V_y: positive = left (see coordinate system above).
    - Shear force V_z: positive = up (see coordinate system above).
    - Torsion M_x: positive = from y to z (twisting around x-axis, see coordinate system above).
    - Bending moment M_y: positive = from z to x (rotation around y-axis, see coordinate system above).
    - Bending moment M_z: positive = from x to y (rotation around z-axis, see coordinate system above).

    Attributes
    ----------
    result_on : ResultOn
        Specifies the object type where the result is located.
        - "On beam": Result on a StructuralCurveMember (requires `member` to be set)
        - "On rib": Result on a StructuralCurveMemberRib (requires `member_rib` to be set)
    member : str
        Reference to the name of the 1D member (StructuralCurveMember).
        Required if and only if result_on = ResultOn.ON_BEAM.
    member_rib : str
        Reference to the name of the 1D member rib (StructuralCurveMemberRib).
        Required if and only if result_on = ResultOn.ON_RIB.
    result_for : ResultFor
        Specifies the source type of the internal force result.
        - "Load case": Result from a single load case (requires `load_case` to be set)
        - "Load combination": Result from a load combination (requires `load_combination` to be set)
    load_case : str
        Reference to the name of the load case (StructuralLoadCase).
        Required if and only if result_for = ResultFor.LOAD_CASE.
    load_combination : str
        Reference to the name of the load combination (StructuralLoadCombination).
        Required if and only if result_for = ResultFor.LOAD_COMBINATION.
    combination_key : str, optional
        Exact combination formula string for envelope or code-based combinations.
        Example: "1.35*LC1+1.5*LC2". Only applicable when result_for = ResultFor.LOAD_COMBINATION.
    section_at : M
        X-coordinate on the beam (distance from start node) where the result is located [m].
    index : int
        Sequential index of the section on the beam, starting at 1 and proceeding from start to end.
        Helps distinguish values on left versus right side of a section at the same location.
    n : KN
        Normal force value [kN].
    vy : KN
        Shear force in Y-axis direction [kN].
    vz : KN
        Shear force in Z-axis direction [kN].
    mx : KNM
        Torsional moment around X-axis [kNm].
    my : KNM
        Bending moment around Y-axis [kNm].
    mz : KNM
        Bending moment around Z-axis [kNm].

    Raises
    ------
    ValueError
        If result_on = ResultOn.ON_BEAM but member is empty.
        If result_on = ResultOn.ON_RIB but member_rib is empty.
        If result_for = ResultFor.LOAD_CASE but load_case is empty.
        If result_for = ResultFor.LOAD_COMBINATION but load_combination is empty.
        If index < 1.
    """

    result_on: ResultOn
    result_for: ResultFor
    member: str = ""
    member_rib: str = ""
    load_case: str = ""
    load_combination: str = ""
    combination_key: str = ""
    section_at: M = 0
    index: int = 1
    n: KN = 0
    vy: KN = 0
    vz: KN = 0
    mx: KNM = 0
    my: KNM = 0
    mz: KNM = 0

    def __post_init__(self) -> None:
        """Validate conditional requirements based on SAF specification.

        Raises
        ------
        ValueError
            If conditional requirements are not satisfied.
        """
        if self.result_on == ResultOn.ON_BEAM and not self.member:
            raise ValueError("member must be specified when result_on = ResultOn.ON_BEAM")
        if self.result_on == ResultOn.ON_RIB and not self.member_rib:
            raise ValueError("member_rib must be specified when result_on = ResultOn.ON_RIB")
        if self.result_for == ResultFor.LOAD_CASE and not self.load_case:
            raise ValueError("load_case must be specified when result_for = ResultFor.LOAD_CASE")
        if self.result_for == ResultFor.LOAD_COMBINATION and not self.load_combination:
            raise ValueError("load_combination must be specified when result_for = ResultFor.LOAD_COMBINATION")
        if self.index < 1:
            raise ValueError("index must be >= 1 (SAF specification requires index >= 1)")



"""Sub-package for steel cross sections."""

from blueprints.structural_sections.steel.profile_definitions.chs_profile import CHSProfile
from blueprints.structural_sections.steel.profile_definitions.i_profile import IProfile
from blueprints.structural_sections.steel.profile_definitions.lnp_profile import LNPProfile
from blueprints.structural_sections.steel.profile_definitions.rhs_profile import RHSProfile
from blueprints.structural_sections.steel.profile_definitions.strip_profile import StripProfile

__all__ = [
    "CHSProfile",
    "IProfile",
    "LNPProfile",
    "RHSProfile",
    "StripProfile",
]



"""Report builder for LaTeX documents.

This module provides functionality for creating structured LaTeX reports programmatically.
The Report class offers a fluent API for building documents with headings, paragraphs,
equations, tables, figures, and lists. Reports can be exported to LaTeX format for
compilation with pdflatex, compiled directly to PDF, or converted to Word documents.

Key Features:
    - Fluent API with method chaining for easy document construction
    - Support for mathematical equations using LaTeX syntax
    - Integration with Blueprints Formula objects
    - Table and figure insertion with customizable formatting
    - Nested bulleted and numbered lists
    - Export to LaTeX (.tex), PDF (.pdf), and Word (.docx) formats
    - Multi-language support through translation

Developer notes:
    The LaTeX styling is designed to match Word document styling as closely as possible.
    Changes to LaTeX output should be reflected in _report_to_word.py converter.

"""

import subprocess
import tempfile
from collections.abc import Sequence
from dataclasses import dataclass, field
from io import BytesIO
from pathlib import Path
from typing import Any, Literal, Self

from blueprints.codes.formula import Formula
from blueprints.utils.language.translate import LatexTranslator
from blueprints.utils.report._report_to_word import _ReportToWordConverter


@dataclass
class Report:
    r"""Report builder for creating structured documents with standardized formatting.

    Check our docs for examples of usage.

    Parameters
    ----------
    title : str, optional
        The title of the report.

    Examples
    --------
    >>> report = Report(title="Sample Report")
    >>> report.add_heading("Introduction")
    >>> report.add_heading("Background", level=2)
    >>> report.add_heading("Details", level=3)
    >>> report.add_paragraph("This is normal text.")
    >>> report.add_paragraph("This is bold text with newline after.", bold=True).add_newline()
    >>> report.add_paragraph("This is italic text with 4 newlines after.", italic=True).add_newline(n=4)
    >>> report.add_paragraph("This is bold and italic text.", bold=True, italic=True)
    >>> report.add_newline()
    >>> report.add_equation("E=mc^2", tag="3.14")
    >>> report.add_paragraph("Before inline equation:", italic=True).add_equation(r"\frac{a}{b}", inline=True).add_paragraph(
    ...     " and after inline equation.", bold=True
    ... ).add_newline()
    >>> report.add_equation(r"e^{i \pi} + 1 = 0", inline=True).add_paragraph("inline can be at start of text.").add_newline()
    >>> report.add_paragraph("Or at the end of text", bold=True).add_equation(r"\int_a^b f(x) dx", inline=True).add_newline(n=2)
    >>> report.add_paragraph("Equations can also be $a^2 + b^2 = c^2$ inline in the add text method.").add_newline()
    >>> report.add_table(
    ...     headers=["Parameter", "Value", "Unit"], rows=[[r"\text{Length}", "10", r"\text{m}"], [r"\text{Density}", "500", r"\text{kg/$m^3$}"]]
    ... )
    >>> report.add_figure(r"tomato.png", width=0.2)  # needs the tomato.png file in working directory
    >>> report.add_list(["First item", "Second item"], style="numbered")
    >>> report.add_list(["Layer one", ["Layer two", ["Layer three", ["Layer four"]]]], style="numbered")
    >>> report.add_list(["Bullet one", "Bullet two"], style="bulleted")
    >>> latex_document = report.to_latex()
    >>> print(latex_document)  # prints the complete LaTeX document string, which can be compiled with pdflatex in for example Overleaf.
    """

    title: str = ""
    content: str = field(default="", init=False)

    def add_paragraph(self, text: str, bold: bool = False, italic: bool = False) -> Self:
        r"""Add text with optional bold and italic formatting.

        Parameters
        ----------
        text : str
            The text content of the paragraph.
        bold : bool, optional
            Whether to format the text in bold.
        italic : bool, optional
            Whether to format the text in italics.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> report.add_paragraph("This is regular text")
        >>> report.add_paragraph("This is bold text", bold=True)
        >>> report.add_paragraph("This is bold and italic", bold=True, italic=True)
        """
        if bold and italic:
            self.content += rf"\textbf{{\textit{{{text}}}}}"
        elif bold:
            self.content += rf"\textbf{{{text}}}"
        elif italic:
            self.content += rf"\textit{{{text}}}"
        else:
            self.content += rf"\txt{{{text}}}"

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_equation(
        self,
        equation: str,
        tag: str | None = None,
        inline: bool = False,
    ) -> Self:
        r"""Add an equation to the report. For adding Blueprints formulas, use add_formula instead.

        Parameters
        ----------
        equation : str
            An equation in LaTeX format.
        tag : str or None, optional
            Tag to label the equation (e.g., "6.83", "EN 1992-1-1:2004 6.6n", etc.).
        inline : bool, optional
            Whether to add the equation inline (meaning within text) or as a separate equation block. Default is False.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        When creating a report, you can add equations in different ways:
        >>> report = Report()
        >>> report.add_equation("a^2+b^2=c^2")
        >>> report.add_equation("a^2+b^2=c^2", tag="6.83")
        >>> report.add_equation(r"\\frac{a}{b}", inline=True)

        """
        if inline:
            self.content += r"\txt{ " + rf"${equation}$" + f"{f' ({tag})' if tag else ''}" + r" }"
        elif tag:
            self.content += rf"\begin{{equation}} {equation} \tag{{{tag}}} \end{{equation}}"
        else:
            self.content += rf"\begin{{equation}} {equation} \notag \end{{equation}}"

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_formula(
        self,
        formula: Formula,
        options: Literal["short", "complete", "complete_with_units"] = "complete",
        n: int = 2,
        include_source: bool = True,
        include_formula_number: bool = True,
        inline: bool = False,
    ) -> Self:
        r"""Add a Blueprints formula to the report, for generic equations, use add_equation.

        Parameters
        ----------
        formula : Formula
            Use any Blueprints Formula object.
        options : Literal["short", "complete", "complete_with_units"]
            The representation of the formula to add.
            short - Minimal representation (symbol = result [unit])
            complete - Complete representation (symbol = equation = numeric_equation = result [unit])
            complete_with_units - Complete representation with units (symbol = equation = numeric_equation_with_units [unit] = result [unit])
        n : int, optional
            Number of decimal places for numerical values in the formula (default is 2).
        include_source: bool, optional
            If True, includes the source document in the equation tag. Default is True.
            For example: "EN 1993-1-1:2005" or "EN 1992-1-1:2004".
        include_formula_number: bool, optional
            If True, includes the formula number in the equation tag. Default is True.
            For example: "6.5" or "6.6n".
        inline : bool, optional
            Whether to add the formula inline (meaning within text) or as a separate equation block (default).

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state import formula_6_5
        >>> formula = formula_6_5.Form6Dot5UnityCheckTensileStrength(n_ed=150000, n_t_rd=200000)
        >>> report = Report()
        >>> report.add_formula(formula, options="short")  # Minimal representation
        >>> print(report)
        # report can be converted to formatted LaTeX document with report.to_document()
        >>> print(report.to_latex())
        """
        # Get the desired LaTeX representation from the formula
        latex = formula.latex(n=n)

        # define the equation string based on options
        equation_str: str = ""
        match options.lower():
            case "short":
                equation_str = latex.short
            case "complete":
                equation_str = latex.complete
            case "complete_with_units":
                equation_str = latex.complete_with_units
            case _:
                raise ValueError(f"Invalid option '{options}'. Choose from 'short', 'complete', or 'complete_with_units'.")

        # Build tag from include_source and include_formula_number flags
        tag_parts = []
        if include_source or include_formula_number:
            if include_source:
                tag_parts.append(formula.source_document)
            if include_formula_number:
                tag_parts.append(formula.label)
        tag_str = " ".join(tag_parts).strip()

        return self.add_equation(equation=equation_str, inline=inline, tag=tag_str or None)

    def add_heading(self, text: str, level: int = 1) -> Self:
        """Add a heading to the report.

        Currently, supports levels 1 (section), 2 (subsection), and 3 (subsubsection).

        Parameters
        ----------
        text : str
            The heading text.
        level : int
            The heading level (1 for section, 2 for subsection, 3 for subsubsection). Default is 1.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> report.add_heading("This is a section")
        """
        match level:
            case 1:
                self.content += rf"\section{{{text}}}"
            case 2:
                self.content += rf"\subsection{{{text}}}"
            case 3:
                self.content += rf"\subsubsection{{{text}}}"
            case _:
                raise ValueError(f"Invalid heading level '{level}'. Choose from 1 (section), 2 (subsection), or 3 (subsubsection).")

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_table(
        self,
        headers: list[str],
        rows: list[list[str]],
        centering: bool = True,
    ) -> Self:
        r"""Add a table to the report.

        Parameters
        ----------
        headers : list[str]
            List of column headers.
        rows : list[list[str]]
            List of rows, where each row is a list of cell values.
        centering : bool, optional
            If True, centers the table. Default is True.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> headers = ["Check", "Utilization", "Status"]
        >>> rows = [[r"\\text{Concrete strut capacity}", "0.588", r"\\text{PASS}"], [r"\\text{Torsion moment capacity}", "4.825", r"\\text{FAIL}"]]
        >>> report.add_table(headers, rows)
        """
        # Validate headers and rows match
        num_cols = len(headers)

        if not headers:
            raise ValueError("At least one header is required.")
        if not rows:
            raise ValueError("At least one row is required.")

        for i, row in enumerate(rows):
            if len(row) != num_cols:
                raise ValueError(
                    f"Row {i} has {len(row)} columns but {num_cols} headers were provided. All rows must have the same number of columns as headers."
                )

        col_spec = "l" * num_cols

        # Build header row
        header_row = " & ".join(headers) + r" \\"

        # Build data rows
        data_rows = " ".join([" & ".join(row) + r" \\" for row in rows])

        # Build table
        centering_cmd = r"\centering " if centering else ""
        table = (
            rf"\begin{{table}}[H] {centering_cmd}"
            rf"\begin{{tabular}}{{{col_spec}}} "
            rf"\toprule {header_row} \midrule {data_rows} "
            rf"\bottomrule \end{{tabular}} \end{{table}}"
        )
        self.content += table

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_figure(
        self,
        image_path: str,
        width: float = 0.9,
        caption: str | None = None,
    ) -> Self:
        r"""Adds a figure to the report.

        Parameters
        ----------
        image_path : str
            Path to the image file.
        width : float, optional
            Width specification for the image as ratio of the text width. Default is 0.9.
        caption : str, optional
            Caption text for the figure. Will be displayed below the image. Default is None.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> report.add_figure("path_to_image")
        >>> report.add_figure("path_to_image", width=0.5)
        >>> report.add_figure("plot.png", caption="Results of the analysis")
        """
        # Convert Windows backslashes to forward slashes for LaTeX compatibility
        latex_image_path = image_path.replace("\\", "/")

        # Build the figure environment
        figure_parts = [r"\begin{figure}[H] \centering ", rf"\includegraphics[width={width}\textwidth]{{{latex_image_path}}} "]

        # Add optional caption
        if caption:
            figure_parts.append(rf"\caption{{{caption}}} ")

        figure_parts.append(r"\end{figure}")

        figure = "".join(figure_parts)
        self.content += figure

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_list(self, items: Sequence[Any], style: Literal["bulleted", "numbered"] = "bulleted") -> Self:
        """Add a list to the report, either bulleted or numbered.

        Parameters
        ----------
        items : Sequence[Any]
            List of items to display.
        style : Literal["bulleted", "numbered"], optional
            Style of the list, either 'bulleted' for itemize or 'numbered' for enumerate. Default is 'bulleted'.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> report.add_list(["Item 1", "Item 2", "Item 3"], style="bulleted")
        >>> report.add_list(["First", "Second", "Third"], style="numbered")
        """
        if style.lower() not in ["bulleted", "numbered"]:
            raise ValueError(f"Invalid style '{style}'. Choose 'bulleted' or 'numbered'.")

        def _build_list(item_list: list, depth: int = 0) -> str:
            r"""Recursively build LaTeX environment for nested lists.

            Parameters
            ----------
            item_list : list
                List of items to convert to LaTeX. Items can be strings or nested lists.
            depth : int, optional
                Current nesting depth (used for recursion tracking). Default is 0.

            Returns
            -------
            str
                LaTeX string with \\begin{itemize}/\\begin{enumerate} environment.
            """
            result = r"\begin{itemize} " if style.lower() == "bulleted" else r"\begin{enumerate} "
            for item in item_list:
                if isinstance(item, list):
                    result += _build_list(item, depth + 1)
                else:
                    result += rf"\item {item} "
            result += r"\end{itemize} " if style.lower() == "bulleted" else r"\end{enumerate} "
            return result

        self.content += _build_list(list(items))

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def add_newline(self, n: int = 1) -> Self:
        """Add one or more newlines to separate content.

        Useful for adding vertical spacing between paragraphs, equations, or other elements.

        Parameters
        ----------
        n : int, optional
            Number of newlines to add. Default is 1.

        Returns
        -------
        Report
            Returns self for method chaining.

        Examples
        --------
        >>> report = Report()
        >>> report.add_newline()
        """
        self.content += r"\newline" * n

        # Add a newline for visual separation
        self.content += "\n"

        return self

    def __add__(self, other: "Report") -> "Report":
        """Combine two reports into a new report.

        The resulting report will have the title of the first (left) report
        and the combined content of both reports.

        Parameters
        ----------
        other : Report
            The report to add to this one.

        Returns
        -------
        Report
            A new Report with combined content.

        Raises
        ------
        TypeError
            If the other object is not a Report instance.

        Examples
        --------
        >>> report1 = Report(title="Part 1")
        >>> report1.add_heading("Introduction")
        Report(title="Part 1", sections=1, subsections=0, equations=0, tables=0, figures=0, lists=0, chars=26)
        >>> report2 = Report(title="Part 2")
        >>> report2.add_heading("Conclusion")
        Report(title="Part 2", sections=1, subsections=0, equations=0, tables=0, figures=0, lists=0, chars=23)
        >>> combined = report1 + report2
        >>> combined.title
        'Part 1'
        """
        if not isinstance(other, Report):
            raise TypeError(f"unsupported operand type(s) for +: 'Report' and '{type(other).__name__}'")
        result = Report(title=self.title)
        result.content = self.content + other.content
        return result

    def __repr__(self) -> str:
        """Return a concise representation showing report structure and content summary."""
        sections = self.content.count(r"\section{")
        subsections = self.content.count(r"\subsection{")
        equations = self.content.count(r"\begin{equation}")
        tables = self.content.count(r"\begin{table}")
        figures = self.content.count(r"\begin{figure}")
        lists = self.content.count(r"\begin{itemize}") + self.content.count(r"\begin{enumerate}")

        title_str = f'title="{self.title}"'
        stats = (
            f"sections={sections}, subsections={subsections}, "
            f"equations={equations}, tables={tables}, figures={figures}, "
            f"lists={lists}, chars={len(self.content)}"
        )

        return f"LatexReport({title_str}, {stats})"

    def __str__(self) -> str:
        """Return a human-readable representation of the report structure and content."""
        sections = self.content.count(r"\section{")
        subsections = self.content.count(r"\subsection{")
        equations = self.content.count(r"\begin{equation}")
        tables = self.content.count(r"\begin{table}")
        figures = self.content.count(r"\begin{figure}")
        lists = self.content.count(r"\begin{itemize}") + self.content.count(r"\begin{enumerate}")

        lines = [
            "=" * 60,
            f"LaTeX Report: {self.title}",
            "=" * 60,
            f"Sections:      {sections}",
            f"Subsections:   {subsections}",
            f"Equations:     {equations}",
            f"Tables:        {tables}",
            f"Figures:       {figures}",
            f"Lists:         {lists}",
            f"Content size:  {len(self.content)} characters",
            "=" * 60,
            "Use .to_latex() to generate the full LaTeX document,",
            ".to_pdf() to compile to PDF (requires pdflatex),",
            "or .to_word() to convert to a Word document.",
            "=" * 60,
        ]

        return "\n".join(lines)

    def to_latex(self, path: str | Path | None = None, language: str = "en") -> str | None:
        """Generate a complete LaTeX document with proper preamble and structure.

        You could compile the output with pdflatex in for example Overleaf.

        Parameters
        ----------
        path : str | Path | None, optional
            The destination for the LaTeX document:
            - str or Path: File path where the .tex file will be saved
            - None: Return the document as a string (default)
        language : str, optional
            Language code for localization, full list on https://docs.cloud.google.com/translate/docs/languages
            Warning: only English is officially supported in Blueprints (default is "en" for English).

        Returns
        -------
        str | None
            If path is None, returns the LaTeX document as a string.
            If path is provided (str or Path), returns None after saving to file.

        Examples
        --------
        Get LaTeX as a string:

        >>> report = Report(title="My Report")
        >>> report.add_heading("Introduction")
        >>> report.add_paragraph("Some text")
        >>> latex_doc = report.to_latex()

        Save directly to a file:

        >>> report.to_latex("report.tex")

        Save using pathlib.Path:

        >>> from pathlib import Path
        >>> report.to_latex(Path("report.tex"))
        """
        # Build the preamble with styling to match Word document converter (pdflatex compatible)
        preamble = (
            r"\documentclass[11pt]{article}" + "\n"
            # Required packages
            r"\usepackage{amsmath}" + "\n"  # Advanced math environments and symbols
            r"\usepackage{booktabs}" + "\n"  # Professional-looking tables with \toprule, \midrule, \bottomrule
            r"\usepackage{float}" + "\n"  # Improved float handling
            r"\usepackage{geometry}" + "\n"  # Page layout and margins
            r"\usepackage{graphicx}" + "\n"  # Include images and graphics
            r"\usepackage{icomma}" + "\n"  # Proper comma handling in numbers
            r"\usepackage{setspace}" + "\n"  # Line spacing control
            r"\usepackage{xcolor}" + "\n"  # Color definitions and usage
            r"\usepackage{titlesec}" + "\n"  # Customize section titles
            r"\usepackage{helvet}" + "\n"  # Helvetica font family (sans-serif)
            r"\usepackage[T1]{fontenc}" + "\n"  # Better font encoding for special characters
            r"\usepackage{enumitem}" + "\n"  # Enhanced list customization
            # Page setup
            r"\geometry{a4paper, margin=1in}" + "\n"  # A4 paper with 1-inch margins
            r"\setstretch{1.3}" + "\n"  # Line spacing factor
            "\n"
            # Custom commands
            r"\newcommand{\txt}[1]{#1}" + "\n"  # Simple text wrapper command
            # Spacing configuration
            r"\setlength{\parskip}{0pt}" + "\n"  # No extra space between paragraphs
            r"\setlength{\abovedisplayskip}{12pt}" + "\n"  # Space above equations
            r"\setlength{\belowdisplayskip}{12pt}" + "\n"  # Space below equations
            r"\setlist{nosep}" + "\n"  # Remove extra spacing in lists
            "\n"
            # Color definitions
            r"\definecolor{blueprintblue}{RGB}{0,40,85}" + "\n"  # Custom blue color (0, 40, 85)
            "\n"
            # Title formatting
            r"\makeatletter" + "\n"  # Access internal LaTeX commands
            r"\renewcommand{\maketitle}{%" + "\n"  # Redefine \maketitle command
            r"    \begin{center}%" + "\n"  # Center the title
            r"        {\sffamily\fontsize{18}{19}\selectfont\bfseries\color{blueprintblue}\@title}%" + "\n"  # 18pt, bold, blue, sans-serif title
            r"        \vspace{4pt}%" + "\n"  # 4pt vertical space after title
            r"    \end{center}%" + "\n"
            r"}" + "\n"
            r"\makeatother" + "\n"  # Restore @ character behavior
            "\n"
            # Section formatting
            r"\titleformat{\section}" + "\n"  # Section heading format
            r"    {\sffamily\fontsize{14}{15}\selectfont\bfseries\color{blueprintblue}}" + "\n"  # 14pt, bold, blue, sans-serif
            r"    {\thesection}{1em}{}" + "\n"  # Section number with 1em space
            r"\titlespacing*{\section}{0pt}{8pt}{4pt}" + "\n"  # Spacing: left, before, after
            "\n"
            # Subsection formatting
            r"\titleformat{\subsection}" + "\n"  # Subsection heading format
            r"    {\sffamily\fontsize{12}{13}\selectfont\bfseries\color{blueprintblue}}" + "\n"  # 12pt, bold, blue, sans-serif
            r"    {\thesubsection}{1em}{}" + "\n"  # Subsection number with 1em space
            r"\titlespacing*{\subsection}{0pt}{8pt}{4pt}" + "\n"  # Spacing: left, before, after
            "\n"
            # Subsubsection formatting
            r"\titleformat{\subsubsection}" + "\n"  # Subsubsection heading format
            r"    {\sffamily\fontsize{12}{13}\selectfont\bfseries\color{blueprintblue}}" + "\n"  # 12pt, bold, blue, sans-serif
            r"    {\thesubsubsection}{1em}{}" + "\n"  # Subsubsection number with 1em space
            r"\titlespacing*{\subsubsection}{0pt}{4pt}{0pt}" + "\n"  # Spacing: left, before, after
            "\n"
            # Paragraph formatting
            r"\parindent 0in" + "\n"  # No paragraph indentation
            # Begin document
            r"\begin{document}" + "\n"
            rf"\title{{{self.title}}}" + "\n"  # Set document title
            r"\date{}" + "\n"  # No date displayed
            r"\maketitle" + "\n"  # Generate the title
        )

        latex = preamble + self.content + r"\end{document}"
        if language != "en":
            # Translate content to the specified language
            latex = LatexTranslator(original_text=latex, destination_language=language).text

        # If path is provided, save to file and return None
        if path is not None:
            # Convert Path to str if needed
            file_path = str(path) if isinstance(path, Path) else path
            Path(file_path).write_text(latex, encoding="utf-8")
            return None

        # Return LaTeX string
        return latex

    def to_word(self, path: str | Path | BytesIO | None = None, language: str = "en") -> bytes | None:  # pragma: no cover
        """Convert the LaTeX report to a Word document.

        This method uses the ReportToWordConverter to convert the LaTeX content
        of the report into a Word document format. The output can be saved to a file,
        written to a BytesIO object, or returned as bytes.

        Parameters
        ----------
        path : str | Path | BytesIO | None, optional
            The destination for the Word document:
            - str or Path: File path where the document will be saved, for example 'report.docx'. Remember to use .docx extension.
            - BytesIO: Buffer to write the document to (in-memory)
                ```python
                from io import BytesIO

                buffer = BytesIO()
                report.to_word(buffer)
                docx_bytes = buffer.getvalue()
                ```
            - None: Return the document as bytes (default)
        language : str, optional
            Language code for localization, full list on https://docs.cloud.google.com/translate/docs/languages
            Warning: only English is officially supported in Blueprints (default is "en" for English).

        Returns
        -------
        bytes | None
            If path is None, returns the Word document as bytes.
            If path is provided (str, Path, or BytesIO), returns None.

        Examples
        --------
        Save to a file path:

        >>> report = Report(title="My Report")
        >>> report.add_heading("Introduction")
        >>> report.add_paragraph("Some text")
        >>> report.to_word("report.docx")  # Save to file

        Save to a pathlib.Path:

        >>> from pathlib import Path
        >>> report.to_word(Path("report.docx"))

        Write to a BytesIO object for in-memory processing:

        >>> from io import BytesIO
        >>> buffer = BytesIO()
        >>> report.to_word(buffer)
        >>> docx_bytes = buffer.getvalue()

        Get bytes directly (useful for streaming, email attachments, etc.):

        >>> docx_bytes = report.to_word()
        >>> # Can now send as email attachment or stream over HTTP
        """
        latex_content = self.to_latex(language=language)
        converter = _ReportToWordConverter(latex_content)
        if converter.document:
            if path is None:
                # Return bytes directly
                buffer = BytesIO()
                converter.document.save(buffer)
                buffer.seek(0)
                return buffer.getvalue()
            # Save to file path or BytesIO object
            # Convert Path to str for compatibility with python-docx
            if isinstance(path, Path):
                converter.document.save(str(path))
            else:
                converter.document.save(path)
            return None
        return None

    def to_pdf(self, path: str | Path | None = None, language: str = "en", cleanup: bool = True) -> bytes | None:  # noqa: C901
        """Generate a PDF document by compiling LaTeX with pdflatex.

        This method generates LaTeX content using to_latex(), compiles it with pdflatex,
        and returns or saves the resulting PDF. Requires pdflatex to be installed and
        available in the system PATH. Can be downloaded from a LaTeX distribution such as TeX Live or MiKTeX.

        Parameters
        ----------
        path : str | Path | None, optional
            The destination for the PDF document:
            - str or Path: File path where the .pdf file will be saved
            - None: Return the PDF as bytes (default)
        language : str, optional
            Language code for localization, full list on https://docs.cloud.google.com/translate/docs/languages
            Warning: only English is officially supported in Blueprints (default is "en" for English).
        cleanup : bool, optional
            Whether to remove temporary LaTeX files after compilation. Default is True.

        Returns
        -------
        bytes | None
            If path is None, returns the PDF document as bytes.
            If path is provided (str or Path), returns None after saving to file.

        Raises
        ------
        RuntimeError
            If pdflatex is not found or compilation fails.

        Examples
        --------
        Get PDF as bytes:

        >>> report = Report(title="My Report")
        >>> report.add_heading("Introduction")
        >>> report.add_paragraph("Some text")
        >>> pdf_bytes = report.to_pdf()

        Save directly to a file:

        >>> report.to_pdf("report.pdf")

        Save using pathlib.Path:

        >>> from pathlib import Path
        >>> report.to_pdf(Path("report.pdf"))

        Keep temporary LaTeX files for debugging:

        >>> report.to_pdf("report.pdf", cleanup=False)
        """
        # Check if pdflatex is available
        try:
            subprocess.run(
                ["pdflatex", "--version"],
                capture_output=True,
                check=True,
                timeout=10,
            )
        except (subprocess.CalledProcessError, FileNotFoundError, subprocess.TimeoutExpired) as e:
            raise RuntimeError(
                "pdflatex is not installed or not found in PATH. Please install a LaTeX distribution (e.g., MiKTeX, TeX Live) that includes pdflatex."
                " Go to https://miktex.org/download or https://www.tug.org/texlive/ for more information."
            ) from e

        # Generate LaTeX content
        latex_content = self.to_latex(language=language)
        assert latex_content is not None  # to_latex returns str when path is None

        # Create temporary directory for compilation
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)
            tex_file = tmpdir_path / "report.tex"
            pdf_file = tmpdir_path / "report.pdf"

            # Write LaTeX content to temporary file
            tex_file.write_text(latex_content, encoding="utf-8")

            # Run pdflatex twice for proper references and table of contents
            for _ in range(2):
                result = subprocess.run(
                    ["pdflatex", "-interaction=nonstopmode", "-halt-on-error", str(tex_file)],
                    check=False,
                    cwd=tmpdir,
                    capture_output=True,
                    text=True,
                    timeout=60,
                )

                if result.returncode != 0:
                    # Extract error information from log
                    error_msg = "pdflatex compilation failed.\n"
                    if result.stdout:  # pragma: no cover
                        # Find the first error in the output
                        lines = result.stdout.split("\n")
                        for i, line in enumerate(lines):
                            if line.startswith("!"):
                                error_msg += "\n".join(lines[i : i + 5])
                                break
                    raise RuntimeError(error_msg)

            # Check if PDF was created
            if not pdf_file.exists():
                raise RuntimeError("PDF file was not created by pdflatex.")  # pragma: no cover

            # Read the PDF content
            pdf_content = pdf_file.read_bytes()

            # If path is provided, save to file
            if path is not None:
                output_path = Path(path) if isinstance(path, str) else path
                output_path.write_bytes(pdf_content)

                # Optionally copy auxiliary files for debugging
                if not cleanup:  # pragma: no cover
                    aux_files = [".aux", ".log", ".out"]
                    base_name = output_path.stem
                    output_dir = output_path.parent
                    for ext in aux_files:
                        aux_file = tmpdir_path / f"report{ext}"
                        if aux_file.exists():
                            (output_dir / f"{base_name}{ext}").write_bytes(aux_file.read_bytes())

                return None

            # Return PDF as bytes
            return pdf_content



"""Formula 6.7 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot7DesignUltimateResistanceNetCrossSection(Formula):
    r"""Class representing formula 6.7 for the calculation of [$N_{t,Rd}$]."""

    label = "6.7"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a_net: MM2,
        f_u: MPA,
        gamma_m2: DIMENSIONLESS,
    ) -> None:
        r"""[$N_{u,Rd}$] Calculation of the design tension resistance [$N$].

        EN 1993-1-1:2005 art.6.2.3(2) - Formula (6.7)

        Parameters
        ----------
        a_net : MM2
            [$A_{net}$] Net cross-sectional area at holes for fasteners [$mm^2$].
        f_u : MPA
            [$f_u$] Ultimate tensile strength of the material [$N/mm^2$].
        gamma_m2 : DIMENSIONLESS
            [$\gamma_{M2}$] Partial safety factor for resistance of cross-sections in tension to fracture.
        """
        super().__init__()
        self.a_net = a_net
        self.f_u = f_u
        self.gamma_m2 = gamma_m2

    @staticmethod
    def _evaluate(
        a_net: MM2,
        f_u: MPA,
        gamma_m2: DIMENSIONLESS,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a_net=a_net, f_u=f_u)
        raise_if_less_or_equal_to_zero(gamma_m2=gamma_m2)

        return 0.9 * a_net * f_u / gamma_m2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.7."""
        _equation: str = r"0.9 \cdot \frac{A_{net} \cdot f_u}{\gamma_{M2}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A_{net}": f"{self.a_net:.{n}f}",
                r"f_u": f"{self.f_u:.{n}f}",
                r"\gamma_{M2}": f"{self.gamma_m2:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"N_{u,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="N",
        )



"""UNP Steel Profiles."""

from typing import NamedTuple

from blueprints.structural_sections.steel.profile_definitions.unp_profile import UNPProfile
from blueprints.structural_sections.steel.standard_profiles.utils import StandardProfileMeta
from blueprints.type_alias import MM, PERCENTAGE


class __UNPProfileParameters(NamedTuple):
    """Parameters for defining a UNP profile."""

    name: str
    top_flange_total_width: MM
    top_flange_thickness: MM
    bottom_flange_total_width: MM
    bottom_flange_thickness: MM
    total_height: MM
    web_thickness: MM
    top_root_fillet_radius: MM
    bottom_root_fillet_radius: MM
    top_toe_radius: MM
    bottom_toe_radius: MM
    top_slope: PERCENTAGE
    bottom_slope: PERCENTAGE


UNP_PROFILES_DATABASE = {
    "UNP80": __UNPProfileParameters("UNP80", 45, 8, 45, 8, 80, 6, 8, 8, 4, 4, 8, 8),
    "UNP100": __UNPProfileParameters("UNP100", 50, 8.5, 50, 8.5, 100, 6, 8.5, 8.5, 4.5, 4.5, 8, 8),
    "UNP120": __UNPProfileParameters("UNP120", 55, 9, 55, 9, 120, 7, 9, 9, 4.5, 4.5, 8, 8),
    "UNP140": __UNPProfileParameters("UNP140", 60, 10, 60, 10, 140, 7, 10, 10, 5, 5, 8, 8),
    "UNP160": __UNPProfileParameters("UNP160", 65, 10.5, 65, 10.5, 160, 7.5, 10.5, 10.5, 5.5, 5.5, 8, 8),
    "UNP180": __UNPProfileParameters("UNP180", 70, 11, 70, 11, 180, 8, 11, 11, 5.5, 5.5, 8, 8),
    "UNP200": __UNPProfileParameters("UNP200", 75, 11.5, 75, 11.5, 200, 8.5, 11.5, 11.5, 6, 6, 8, 8),
    "UNP220": __UNPProfileParameters("UNP220", 80, 12.5, 80, 12.5, 220, 9, 12.5, 12.5, 6.5, 6.5, 8, 8),
    "UNP240": __UNPProfileParameters("UNP240", 85, 13, 85, 13, 240, 9.5, 13, 13, 6.5, 6.5, 8, 8),
    "UNP260": __UNPProfileParameters("UNP260", 90, 14, 90, 14, 260, 10, 14, 14, 7, 7, 8, 8),
    "UNP280": __UNPProfileParameters("UNP280", 95, 15, 95, 15, 280, 10, 15, 15, 7.5, 7.5, 8, 8),
    "UNP300": __UNPProfileParameters("UNP300", 100, 16, 100, 16, 300, 10, 16, 16, 8, 8, 8, 8),
    "UNP320": __UNPProfileParameters("UNP320", 100, 17.5, 100, 17.5, 320, 14, 17.5, 17.5, 8.75, 8.75, 5, 5),
    "UNP350": __UNPProfileParameters("UNP350", 100, 16, 100, 16, 350, 14, 16, 16, 8, 8, 5, 5),
    "UNP380": __UNPProfileParameters("UNP380", 102, 16, 102, 16, 380, 13.5, 16, 16, 8, 8, 5, 5),
    "UNP400": __UNPProfileParameters("UNP400", 110, 18, 110, 18, 400, 14, 18, 18, 9, 9, 5, 5),
}


class UNP(metaclass=StandardProfileMeta):
    """Geometrical representation of UNP profiles.

    This class provides access to standard UNP profiles from a predefined database.
    Profiles can be accessed as class attributes using their standardized names.
    Each accessed profile returns a UNPProfile instance.

    Usage example
    -------------
        >>> profile = UNP.UNP200
    """

    _factory = UNPProfile
    """Factory class for creating standard UNP profiles."""
    _database = UNP_PROFILES_DATABASE
    """Database of standard UNP profile parameters."""



"""Utility functions and (meta)classes for standard steel cross-section profiles."""

from __future__ import annotations

from collections.abc import Callable
from typing import NamedTuple, Protocol

from blueprints.structural_sections._profile import Profile


class StandardProfileProtocol(Protocol):
    """Protocol for standard profile classes."""

    _factory: Callable[..., Profile]
    """Factory class for creating standard profiles."""
    _database: dict[str, NamedTuple]
    """Database of standard profiles."""


class StandardProfileMeta(type):
    """Metaclass for standard profile classes to enable dynamic attribute access."""

    def __getattr__(cls: StandardProfileProtocol, name: str) -> Profile:
        """Get a profile by its name from the class database.

        Parameters
        ----------
        name : str
            The name of the profile to retrieve.

        Returns
        -------
        Profile
            An instance of the profile corresponding to the given name.

        Raises
        ------
        AttributeError
            If the profile name does not exist in the database.
        """
        try:
            profile = cls._database[name]
        except KeyError as e:
            raise AttributeError(f"Profile '{name}' does not exist in database.") from e
        return cls._factory(**profile._asdict())



"""Test suite for IProfile."""

import matplotlib as mpl

mpl.use("Agg")

from unittest.mock import MagicMock

import pytest
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from blueprints.structural_sections.steel.profile_definitions.i_profile import IProfile
from blueprints.validations import NegativeValueError


class TestIProfile:
    """Test suite for IProfile."""

    def test_alias(self, h_profile: IProfile) -> None:
        """Test the alias of the I-profile."""
        expected_alias = "HEB360"
        assert h_profile.name == expected_alias

    def test_maximum_element_thickness(self, h_profile: IProfile) -> None:
        """Test the maximum element thickness of the I-profile."""
        expected_max_thickness = 22.5  # mm
        assert pytest.approx(h_profile.max_thickness, rel=1e-6) == expected_max_thickness

    def test_steel_volume_per_meter(self, h_profile: IProfile) -> None:
        """Test the steel volume per meter."""
        expected_volume = 1.806e-2  # m³/m
        assert pytest.approx(h_profile.volume_per_meter, rel=1e-2) == expected_volume

    def test_steel_area(self, h_profile: IProfile) -> None:
        """Test the steel cross-sectional area."""
        expected_area = 1.806e4  # mm²
        assert pytest.approx(h_profile.area, rel=1e-2) == expected_area

    @pytest.mark.slow
    def test_plot_ipe_profile(self, ipe_profile: IProfile) -> None:
        """Test the plot method for an IPE profile (ensure it runs without errors)."""
        fig: Figure = ipe_profile.plot()
        assert isinstance(fig, plt.Figure)

    @pytest.mark.slow
    def test_plot_h_profile(self, h_profile: IProfile) -> None:
        """Test the plot method for an H-shaped profile (ensure it runs without errors)."""
        fig: Figure = h_profile.plot()
        assert isinstance(fig, plt.Figure)

    def test_plot_mocked(self, h_profile: IProfile, mock_section_properties: MagicMock) -> None:  # noqa: ARG002
        """Test the plotting of the I-profile shapes with mocked section properties."""
        fig: Figure = h_profile.plot()
        assert isinstance(fig, plt.Figure)

    def test_geometry(self, h_profile: IProfile) -> None:
        """Test the geometry of the I profile."""
        expected_geometry = h_profile._geometry()  # noqa: SLF001
        assert expected_geometry is not None

    def test_immutability(self, h_profile: IProfile) -> None:
        """Test that the IProfile dataclass is immutable."""
        with pytest.raises(AttributeError):
            h_profile.name = "New Name"  # type: ignore[misc]

    def test_transform(self, h_profile: IProfile) -> None:
        """Test the transform method of the I profile."""
        transformed_profile = h_profile.transform(horizontal_offset=1000, vertical_offset=500, rotation=90)
        assert transformed_profile is not None
        assert isinstance(transformed_profile, IProfile)
        assert pytest.approx(transformed_profile.centroid.x, rel=1e-6) == h_profile.centroid.x + 1000
        assert pytest.approx(transformed_profile.centroid.y, rel=1e-6) == h_profile.centroid.y + 500
        assert pytest.approx(transformed_profile.profile_height, rel=1e-6) == h_profile.profile_width

    def test_with_corrosion_negative_value(self, h_profile: IProfile) -> None:
        """Test that negative corrosion value raises NegativeValueError."""
        with pytest.raises(NegativeValueError, match=r"corrosion"):
            h_profile.with_corrosion(corrosion=-1)

    def test_with_corrosion_zero_value(self, h_profile: IProfile) -> None:
        """Test that zero corrosion returns the same profile instance."""
        corroded_profile = h_profile.with_corrosion(corrosion=0)
        # Zero corrosion should return the same instance
        assert corroded_profile is h_profile
        assert corroded_profile.name == h_profile.name

    def test_with_corrosion_valid_positive_value(self, h_profile: IProfile) -> None:
        """Test applying a valid positive corrosion value."""
        corrosion = 2.0  # mm
        corroded_profile = h_profile.with_corrosion(corrosion=corrosion)

        # Check that a new instance is returned
        assert corroded_profile is not h_profile

        # Check dimensions are reduced by 2 * corrosion (both sides)
        assert pytest.approx(corroded_profile.top_flange_width, rel=1e-6) == h_profile.top_flange_width - corrosion * 2
        assert pytest.approx(corroded_profile.top_flange_thickness, rel=1e-6) == h_profile.top_flange_thickness - corrosion * 2
        assert pytest.approx(corroded_profile.bottom_flange_width, rel=1e-6) == h_profile.bottom_flange_width - corrosion * 2
        assert pytest.approx(corroded_profile.bottom_flange_thickness, rel=1e-6) == h_profile.bottom_flange_thickness - corrosion * 2
        assert pytest.approx(corroded_profile.total_height, rel=1e-6) == h_profile.total_height - corrosion * 2
        assert pytest.approx(corroded_profile.web_thickness, rel=1e-6) == h_profile.web_thickness - corrosion * 2

        # Check radius values are increased by corrosion
        assert corroded_profile.top_radius == h_profile.top_radius + corrosion
        assert corroded_profile.bottom_radius == h_profile.bottom_radius + corrosion

        # Check name is updated with corrosion info
        expected_name = "HEB360 (corrosion: 2.0 mm)"
        assert corroded_profile.name == expected_name

    def test_with_corrosion_fully_corroded_profile(self, h_profile: IProfile) -> None:
        """Test that applying corrosion that fully corrodes the profile raises ValueError."""
        # HEB360 has web_thickness = 12.5 mm and flange_thickness = 22.5 mm
        # Apply a corrosion large enough to fully corrode the web
        corrosion = 10.0  # mm - this will cause web_thickness to be 12.5 - 20 = -7.5 mm
        with pytest.raises(ValueError, match=r"The profile has fully corroded."):
            h_profile.with_corrosion(corrosion=corrosion)

    def test_with_corrosion_existing_corrosion_total_in_name(self, h_profile: IProfile) -> None:
        """Test that applying corrosion to a profile with existing corrosion shows total corrosion in name."""
        # First apply 1 mm corrosion
        first_corrosion = 1.0  # mm
        first_corroded_profile = h_profile.with_corrosion(corrosion=first_corrosion)
        assert first_corroded_profile.name == "HEB360 (corrosion: 1.0 mm)"

        # Then apply another 1.5 mm corrosion
        second_corrosion = 1.5  # mm
        second_corroded_profile = first_corroded_profile.with_corrosion(corrosion=second_corrosion)

        # Check that the name shows total corrosion (1.0 + 1.5 = 2.5 mm)
        expected_name = "HEB360 (corrosion: 2.5 mm)"
        assert second_corroded_profile.name == expected_name

        # Check dimensions reflect total corrosion
        total_corrosion = first_corrosion + second_corrosion
        assert pytest.approx(second_corroded_profile.top_flange_width, rel=1e-6) == h_profile.top_flange_width - total_corrosion * 2
        assert pytest.approx(second_corroded_profile.top_flange_thickness, rel=1e-6) == h_profile.top_flange_thickness - total_corrosion * 2
        assert pytest.approx(second_corroded_profile.bottom_flange_width, rel=1e-6) == h_profile.bottom_flange_width - total_corrosion * 2
        assert pytest.approx(second_corroded_profile.bottom_flange_thickness, rel=1e-6) == h_profile.bottom_flange_thickness - total_corrosion * 2
        assert pytest.approx(second_corroded_profile.total_height, rel=1e-6) == h_profile.total_height - total_corrosion * 2
        assert pytest.approx(second_corroded_profile.web_thickness, rel=1e-6) == h_profile.web_thickness - total_corrosion * 2



"""Module for checking tension force resistance of steel cross-sections."""

from dataclasses import dataclass

from blueprints.checks.check_result import CheckResult
from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state import (
    formula_6_5,
    formula_6_6,
)
from blueprints.codes.formula import Formula
from blueprints.structural_sections.steel.steel_cross_section import SteelCrossSection
from blueprints.type_alias import DIMENSIONLESS, KN
from blueprints.unit_conversion import KN_TO_N
from blueprints.utils.report import Report
from blueprints.validations import NegativeValueError


@dataclass(frozen=True)
class CheckStrengthTensionClass1234:
    """Tension force resistance check for steel cross-sections (class 1,2,3 and 4) based on EN 1993-1-1:2005 art. 6.2.3.

    Coordinate System:
    ```
        z (vertical)
            ↑
            |     x (longitudinal beam direction, into screen)
            |    ↗
            |   /
            |  /
            | /
            |/
      ←-----O
       y (horizontal/side)
    ```

    Parameters
    ----------
    steel_cross_section : SteelCrossSection
        The steel cross-section to check.
    n : KN, optional
        The applied tensile force (positive value), default is 0 kN.
        Will raise an error if a negative value is provided, as this check is only for tension.
    gamma_m0 : DIMENSIONLESS, optional
        Partial safety factor for resistance of cross-sections, default is 1.0.

    Example
    -------
    ```python
    from blueprints.checks import CheckStrengthTensionClass1234
    from blueprints.materials.steel import SteelMaterial, SteelStrengthClass
    from blueprints.structural_sections.steel import SteelCrossSection
    from blueprints.structural_sections.steel.standard_profiles.heb import HEB

    steel_material = SteelMaterial(steel_class=SteelStrengthClass.S355)
    heb_300_profile = HEB.HEB300.with_corrosion(corrosion=1.5)
    n = 10000  # Applied tensile force in kN

    heb_300_s355 = SteelCrossSection(profile=heb_300_profile, material=steel_material)
    calc = CheckStrengthTensionClass1234(steel_cross_section=heb_300_s355, n=n, gamma_m0=1.0)
    calc.report().to_word(path="tension_strength.docx")
    ```

    Raises
    ------
    NegativeValueError
        If the applied force is not tension (i.e., if n < 0).

    """

    steel_cross_section: SteelCrossSection
    n: KN = 0
    gamma_m0: DIMENSIONLESS = 1.0
    name: str = "Tension strength check for steel profiles"

    def __post_init__(self) -> None:
        """Post-initialization to validate input parameters."""
        if self.n < 0:
            raise NegativeValueError(value_name="n (tensile force)", value=self.n)

    @staticmethod
    def source_docs() -> list[str]:
        """List of source document identifiers used for this check.

        Returns
        -------
        list[str]
        """
        return [EN_1993_1_1_2005]

    def plastic_resistance(self) -> Formula:
        """Calculate the tension force plastic resistance of the steel cross-section based on the gross
        cross-sectional area and yield strength (EN 1993-1-1:2005 art. 6.2.3(2a) - Formula (6.6)).

        Returns
        -------
        Formula
            The calculated tension force resistance.
        """
        a = self.steel_cross_section.profile.section_properties().area
        assert a is not None, "Cross-sectional area must be defined for the steel profile."
        f_y = self.steel_cross_section.yield_strength
        return formula_6_6.Form6Dot6DesignPlasticResistanceGrossCrossSection(a=a, f_y=f_y, gamma_m0=self.gamma_m0)

    def tensile_strength_unity_check(self) -> Formula:
        """Calculate the unity check for tensile strength of the steel cross-section based on the applied tensile
        force and the calculated resistance (EN 1993-1-1:2005 art. 6.2.3(1) - Formula (6.5)).

        Returns
        -------
        Formula
            The calculated unity check for tensile strength.
        """
        n_ed = self.n * KN_TO_N
        n_t_rd = self.plastic_resistance()
        return formula_6_5.Form6Dot5UnityCheckTensileStrength(n_ed=n_ed, n_t_rd=n_t_rd)

    def result(self) -> CheckResult:
        """Calculate result of tension force resistance.

        Returns
        -------
        CheckResult
            This is the result of the tension force resistance check, which compares the provided tensile force
            with the calculated resistance. The check is satisfied if the provided tensile force does not exceed
            the resistance.
        """
        return CheckResult.from_comparison(provided=self.n * KN_TO_N, required=self.plastic_resistance())

    def report(self, n: int = 2) -> Report:
        """Returns the report for the tension force check.

        Parameters
        ----------
        n : int, optional
            Number of decimal places for numerical values in the report (default is 2).

        Returns
        -------
        Report
            Full report on the tension force check, including the applied force, calculated resistance,
            unity check, and overall result.
        """
        report = Report("Tension check of steel cross-sections")

        # will not generate a report if no tensile force is applied, as the check is not necessary in that case
        if self.n == 0:
            report.add_paragraph("No tensile force was applied; therefore, no tensile force check is necessary.")
            return report

        # generate report if tensile force is applied
        report.add_paragraph(
            f"Profile {self.steel_cross_section.profile.name} with steel quality {self.steel_cross_section.material.steel_class.name} "
            f"is loaded with a tensile force of {self.n:.{n}f} kN."
        )
        report.add_newline(n=2)

        # resistance
        report.add_paragraph("The resistance is calculated as follows:")
        report.add_formula(self.plastic_resistance(), n=n)
        report.add_newline(n=2)

        # unity check
        report.add_paragraph("The unity check is calculated as follows:")
        report.add_formula(self.tensile_strength_unity_check(), n=n)
        report.add_newline(n=2)

        # add overall result based on the unity check
        if self.result().is_ok:
            report.add_paragraph("The check for tensile force satisfies the requirements.")
        else:
            report.add_paragraph("The check for tensile force does NOT satisfy the requirements.")
        return report



"""Tests for hexagonal profile shapes."""

import numpy as np
import pytest

from blueprints.structural_sections.geometric_profiles.hexagon import HexagonalProfile


class TestHexagonalProfile:
    """Tests for the HexagonalProfile class."""

    def test_area(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the area property of the HexagonalProfile class."""
        expected_area = (3 * np.sqrt(3) / 2) * 50.0**2
        assert hexagonal_profile.area == pytest.approx(expected=expected_area, rel=1e-6)

    def test_polygon(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the geometry property of the HexagonalProfile class."""
        polygon = hexagonal_profile.polygon
        assert polygon.is_valid
        assert polygon.area == pytest.approx(expected=hexagonal_profile.area, rel=1e-4)

    def test_invalid_side_length(self) -> None:
        """Test initialization with an invalid side length value."""
        with pytest.raises(ValueError, match="Side length must be a positive value"):
            HexagonalProfile(name="InvalidHexagon", side_length=-10.0, x=0.0, y=0.0)

    def test_section(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the section object of the HexagonalProfile class."""
        section = hexagonal_profile._section()  # noqa: SLF001
        assert section is not None

    def test_geometry(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the geometry property of the HexagonalProfile class."""
        geometry = hexagonal_profile._geometry()  # noqa: SLF001
        assert geometry is not None

    def test_mesh_settings(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the mesh_settings property of the HexagonalProfile class."""
        mesh_settings = hexagonal_profile.mesh_settings
        assert isinstance(mesh_settings, dict)
        assert "mesh_sizes" in mesh_settings

    def test_apothem(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the apothem property of the HexagonalProfile class."""
        expected_apothem = hexagonal_profile.side_length * np.sqrt(3) / 2
        assert hexagonal_profile.apothem == pytest.approx(expected=expected_apothem, rel=1e-6)

    def test_perimeter(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the perimeter property of the HexagonalProfile class."""
        expected_perimeter = 6 * hexagonal_profile.side_length
        assert hexagonal_profile.perimeter == pytest.approx(expected=expected_perimeter, rel=1e-6)

    def test_parameters_as_dict(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the parameters_as_dict method of the HexagonalProfile class."""
        params = hexagonal_profile.section_properties().asdict()
        assert params

    def test_no_plotter_defined(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test that accessing the plotter property raises an AttributeError if no plotter is defined."""
        with pytest.raises(AttributeError, match=r"No plotter is defined."):
            _ = hexagonal_profile.plotter

    def test_immutability(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test that the HexagonalProfile dataclass is immutable."""
        with pytest.raises(AttributeError):
            hexagonal_profile.name = "New Name"  # type: ignore[misc]

    def test_transform(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the transform method of the HexagonalProfile class."""
        transformed_profile = hexagonal_profile.transform(horizontal_offset=10.0, vertical_offset=20.0, rotation=90.0)
        assert isinstance(transformed_profile, HexagonalProfile)
        assert pytest.approx(transformed_profile.centroid.x, rel=1e-6) == hexagonal_profile.centroid.x + 10.0
        assert pytest.approx(transformed_profile.centroid.y, rel=1e-6) == hexagonal_profile.centroid.y + 20.0
        assert pytest.approx(transformed_profile.profile_height, rel=1e-6) == hexagonal_profile.profile_width

    def test_max_thickness(self, hexagonal_profile: HexagonalProfile) -> None:
        """Test the max_thickness property of the HexagonalProfile class."""
        expected_max_thickness = hexagonal_profile.side_length * np.sqrt(3)
        assert hexagonal_profile.max_thickness == pytest.approx(expected=expected_max_thickness, rel=1e-6)



from blueprints.structural_sections.steel.profile_definitions.rhs_profile import RHSProfile
from blueprints.type_alias import MM

class RHSCF:
    RHSCF40x20x2: RHSProfile
    RHSCF40x20x2_5: RHSProfile
    RHSCF40x20x3: RHSProfile
    RHSCF50x30x2: RHSProfile
    RHSCF50x30x2_5: RHSProfile
    RHSCF50x30x3: RHSProfile
    RHSCF50x30x4: RHSProfile
    RHSCF60x40x2: RHSProfile
    RHSCF60x40x2_5: RHSProfile
    RHSCF60x40x3: RHSProfile
    RHSCF60x40x4: RHSProfile
    RHSCF60x40x5: RHSProfile
    RHSCF70x50x2: RHSProfile
    RHSCF70x50x2_5: RHSProfile
    RHSCF70x50x3: RHSProfile
    RHSCF70x50x4: RHSProfile
    RHSCF70x50x5: RHSProfile
    RHSCF80x40x2: RHSProfile
    RHSCF80x40x2_5: RHSProfile
    RHSCF80x40x3: RHSProfile
    RHSCF80x40x4: RHSProfile
    RHSCF80x40x5: RHSProfile
    RHSCF80x60x2: RHSProfile
    RHSCF80x60x2_5: RHSProfile
    RHSCF80x60x3: RHSProfile
    RHSCF80x60x4: RHSProfile
    RHSCF80x60x5: RHSProfile
    RHSCF90x50x2: RHSProfile
    RHSCF90x50x2_5: RHSProfile
    RHSCF90x50x3: RHSProfile
    RHSCF90x50x4: RHSProfile
    RHSCF90x50x5: RHSProfile
    RHSCF100x40x2_5: RHSProfile
    RHSCF100x40x3: RHSProfile
    RHSCF100x40x4: RHSProfile
    RHSCF100x40x5: RHSProfile
    RHSCF100x50x2_5: RHSProfile
    RHSCF100x50x3: RHSProfile
    RHSCF100x50x4: RHSProfile
    RHSCF100x50x5: RHSProfile
    RHSCF100x50x6: RHSProfile
    RHSCF100x50x6_3: RHSProfile
    RHSCF100x60x2_5: RHSProfile
    RHSCF100x60x3: RHSProfile
    RHSCF100x60x4: RHSProfile
    RHSCF100x60x5: RHSProfile
    RHSCF100x60x6: RHSProfile
    RHSCF100x60x6_3: RHSProfile
    RHSCF100x80x2_5: RHSProfile
    RHSCF100x80x3: RHSProfile
    RHSCF100x80x4: RHSProfile
    RHSCF100x80x5: RHSProfile
    RHSCF100x80x6: RHSProfile
    RHSCF100x80x6_3: RHSProfile
    RHSCF120x60x2_5: RHSProfile
    RHSCF120x60x3: RHSProfile
    RHSCF120x60x4: RHSProfile
    RHSCF120x60x5: RHSProfile
    RHSCF120x60x6: RHSProfile
    RHSCF120x60x6_3: RHSProfile
    RHSCF120x60x8: RHSProfile
    RHSCF120x80x3: RHSProfile
    RHSCF120x80x4: RHSProfile
    RHSCF120x80x5: RHSProfile
    RHSCF120x80x6: RHSProfile
    RHSCF120x80x6_3: RHSProfile
    RHSCF120x80x8: RHSProfile
    RHSCF140x80x4: RHSProfile
    RHSCF140x80x5: RHSProfile
    RHSCF140x80x6: RHSProfile
    RHSCF140x80x6_3: RHSProfile
    RHSCF140x80x8: RHSProfile
    RHSCF150x100x4: RHSProfile
    RHSCF150x100x5: RHSProfile
    RHSCF150x100x6: RHSProfile
    RHSCF150x100x6_3: RHSProfile
    RHSCF150x100x8: RHSProfile
    RHSCF150x100x10: RHSProfile
    RHSCF150x100x12: RHSProfile
    RHSCF150x100x12_5: RHSProfile
    RHSCF160x80x4: RHSProfile
    RHSCF160x80x5: RHSProfile
    RHSCF160x80x6: RHSProfile
    RHSCF160x80x6_3: RHSProfile
    RHSCF160x80x8: RHSProfile
    RHSCF160x80x10: RHSProfile
    RHSCF160x80x12: RHSProfile
    RHSCF160x80x12_5: RHSProfile
    RHSCF180x100x4: RHSProfile
    RHSCF180x100x5: RHSProfile
    RHSCF180x100x6: RHSProfile
    RHSCF180x100x6_3: RHSProfile
    RHSCF180x100x8: RHSProfile
    RHSCF180x100x10: RHSProfile
    RHSCF180x100x12: RHSProfile
    RHSCF180x100x12_5: RHSProfile
    RHSCF200x100x4: RHSProfile
    RHSCF200x100x5: RHSProfile
    RHSCF200x100x6: RHSProfile
    RHSCF200x100x6_3: RHSProfile
    RHSCF200x100x8: RHSProfile
    RHSCF200x100x10: RHSProfile
    RHSCF200x100x12: RHSProfile
    RHSCF200x100x12_5: RHSProfile
    RHSCF200x120x4: RHSProfile
    RHSCF200x120x5: RHSProfile
    RHSCF200x120x6: RHSProfile
    RHSCF200x120x6_3: RHSProfile
    RHSCF200x120x8: RHSProfile
    RHSCF200x120x10: RHSProfile
    RHSCF200x120x12: RHSProfile
    RHSCF200x120x12_5: RHSProfile
    RHSCF250x150x5: RHSProfile
    RHSCF250x150x6: RHSProfile
    RHSCF250x150x6_3: RHSProfile
    RHSCF250x150x8: RHSProfile
    RHSCF250x150x10: RHSProfile
    RHSCF250x150x12: RHSProfile
    RHSCF250x150x12_5: RHSProfile
    RHSCF250x150x16: RHSProfile
    RHSCF260x180x5: RHSProfile
    RHSCF260x180x6_3: RHSProfile
    RHSCF260x180x8: RHSProfile
    RHSCF260x180x10: RHSProfile
    RHSCF260x180x12: RHSProfile
    RHSCF260x180x12_5: RHSProfile
    RHSCF260x180x16: RHSProfile
    RHSCF300x100x6: RHSProfile
    RHSCF300x100x6_3: RHSProfile
    RHSCF300x100x8: RHSProfile
    RHSCF300x100x10: RHSProfile
    RHSCF300x100x12: RHSProfile
    RHSCF300x100x12_5: RHSProfile
    RHSCF300x100x16: RHSProfile
    RHSCF300x150x6: RHSProfile
    RHSCF300x150x6_3: RHSProfile
    RHSCF300x150x8: RHSProfile
    RHSCF300x150x10: RHSProfile
    RHSCF300x150x12: RHSProfile
    RHSCF300x150x12_5: RHSProfile
    RHSCF300x150x16: RHSProfile
    RHSCF300x200x6: RHSProfile
    RHSCF300x200x6_3: RHSProfile
    RHSCF300x200x8: RHSProfile
    RHSCF300x200x10: RHSProfile
    RHSCF300x200x12: RHSProfile
    RHSCF300x200x12_5: RHSProfile
    RHSCF300x200x16: RHSProfile
    RHSCF350x250x6: RHSProfile
    RHSCF350x250x6_3: RHSProfile
    RHSCF350x250x8: RHSProfile
    RHSCF350x250x10: RHSProfile
    RHSCF350x250x12: RHSProfile
    RHSCF350x250x12_5: RHSProfile
    RHSCF350x250x16: RHSProfile
    RHSCF400x200x8: RHSProfile
    RHSCF400x200x12_5: RHSProfile
    RHSCF400x200x16: RHSProfile
    RHSCF400x300x8: RHSProfile
    RHSCF400x300x10: RHSProfile
    RHSCF400x300x12: RHSProfile
    RHSCF400x300x12_5: RHSProfile
    RHSCF400x300x16: RHSProfile

class __RHSCFProfileParameters:
    name: str
    total_height: MM
    total_width: MM
    left_wall_thickness: MM
    right_wall_thickness: MM
    top_wall_thickness: MM
    bottom_wall_thickness: MM
    top_right_outer_radius: MM
    top_left_outer_radius: MM
    bottom_right_outer_radius: MM
    bottom_left_outer_radius: MM
    top_right_inner_radius: MM
    top_left_inner_radius: MM
    bottom_right_inner_radius: MM
    bottom_left_inner_radius: MM

RHSCF_PROFILES_DATABASE: dict[str, __RHSCFProfileParameters]



"""Testing Formula E.4 from EN 1995-1-1:2023."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1995_1_1_2023.appendix_e.formula_e_4 import FormEDot4DistanceToCentroidA2
from blueprints.validations import LessOrEqualToZeroError


class TestFormEDot4DistanceToCentroidA2:
    """Validation for formula E.4 from EN 1995-1-1:2023."""

    @pytest.mark.parametrize(
        ("e_i", "a_i", "gamma_i", "h_i", "expected_result"),
        [
            ([1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], -1.10),
            ([1, 1, 200], [1, 1, 400], [1, 1, 0.2], [1, 1, 0.25], -0.625),
            ([1, 1], [1, 1], [1, 1], [1, 1], 0.50),
            ([1000, 2000], [2000, 2000], [0.5, 0.8], [30, 30], 7.143),
        ],
        ids=["test-3arg", "test-3argb", "test-2arg", "test-2argb"],
    )
    def test_evaluation(self, e_i: list[float], a_i: list[float], gamma_i: list[float], h_i: list[float], expected_result: float) -> None:
        """Test the evaluation of the result."""
        form_e_4 = FormEDot4DistanceToCentroidA2(e_i=e_i, a_i=a_i, gamma_i=gamma_i, h_i=h_i)
        assert form_e_4 == pytest.approx(expected=expected_result, abs=1e-3)

    @pytest.mark.parametrize(
        ("e_i", "a_i", "gamma_i", "h_i", "expectation"),
        [
            ([1, 1, 2], [1, 1, 2], [1, 1, 2], [1, 1, 2], does_not_raise()),
            ([1, 1], [1, 1], [1, 1], [1, 1], does_not_raise()),
            ([1, 1], [1, 1], [1, -1], [1, 1], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1], [1, 1], [1, 1], [1, 0], pytest.raises(LessOrEqualToZeroError)),
            ([-1, 1], [1, 1], [1, 1], [1, 1], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1], [0, 1], [1, 1], [1, 1], pytest.raises(LessOrEqualToZeroError)),
            ([1, 1, 0], [1, 1], [1, 1], [1, 1], pytest.raises(ValueError, match="All input lists must have the same length")),
            (
                [1, 1, 0, 2],
                [1, 1, 2, 3],
                [1, 1, 2, 1],
                [1, 1, 1, 1],
                pytest.raises(ValueError, match="The length of the lists/layers must be either 2 or 3"),
            ),
        ],
        ids=["integer number", "decimal number", "negative_gamma_i", "zero_h_i", "negative_E_i", "zero_A_i", "unequal_lists", "long lists"],
    )
    def test_raise_error_incorrect_args(
        self, e_i: list[float], a_i: list[float], gamma_i: list[float], h_i: list[float], expectation: AbstractContextManager
    ) -> None:
        """Test the evaluation of incorrect arguments."""
        with expectation:
            assert FormEDot4DistanceToCentroidA2(e_i=e_i, a_i=a_i, gamma_i=gamma_i, h_i=h_i) is not None

    @pytest.mark.parametrize(
        ("e_i", "a_i", "gamma_i", "h_i", "rep_short", "rep_long"),
        [
            (
                [1, 1],
                [1, 1],
                [1, 1],
                [1, 1],
                r"\alpha_{2} = 0.50",
                r"\alpha_{2} = \frac{\gamma_1 E_1 A_1 (h_1 + h_2)}{2 \left(\gamma_1 E_1 A_1 + \gamma_2 E_2 A_2\right)}"
                r" = \frac{1.00 \cdot 1.00 \cdot 1.00 (1.00 + 1.00)}{2 \left(1.00 \cdot 1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00\right)}"
                " = 0.50",
            ),
            (
                [1, 1, 200],
                [1, 1, 400],
                [1, 1, 0.2],
                [1, 1, 0.25],
                r"\alpha_{2} = -0.62",
                r"\alpha_{2} = \frac{\gamma_1 E_1 A_1 (h_1 + h_2) - \gamma_3 E_3 A_3 (h_2 + h_3)}"
                r"{2 \left(\gamma_1 E_1 A_1 + \gamma_2 E_2 A_2 + \gamma_3 E_3 A_3\right)}"
                r" = \frac{1.00 \cdot 1.00 \cdot 1.00 (1.00 + 1.00) - 0.20 \cdot 200.00 \cdot 400.00 (1.00 + 0.25)}"
                r"{2 \left(1.00 \cdot 1.00 \cdot 1.00 + 1.00 \cdot 1.00 \cdot 1.00 + 0.20 \cdot 200.00 \cdot 400.00\right)}"
                " = -0.62",
            ),
        ],
        ids=["length_2", "length_3"],
    )
    def test_latex(self, e_i: list[float], a_i: list[float], gamma_i: list[float], h_i: list[float], rep_short: str, rep_long: str) -> None:
        """Test the latex representation of the formula."""
        form_e_4_latex = FormEDot4DistanceToCentroidA2(e_i=e_i, a_i=a_i, gamma_i=gamma_i, h_i=h_i).latex()
        assert form_e_4_latex.complete == rep_long
        assert form_e_4_latex.short == rep_short



"""HEM Steel Profiles."""

from typing import NamedTuple

from blueprints.structural_sections.steel.profile_definitions.i_profile import IProfile
from blueprints.structural_sections.steel.standard_profiles.utils import StandardProfileMeta
from blueprints.type_alias import MM


class __HEMProfileParameters(NamedTuple):
    """Parameters for defining a HEM profile."""

    name: str
    """Name of the HEM profile."""
    top_flange_width: MM
    """Top flange width of the HEM profile."""
    top_flange_thickness: MM
    """Top flange thickness of the HEM profile."""
    bottom_flange_width: MM
    """Bottom flange width of the HEM profile."""
    bottom_flange_thickness: MM
    """Bottom flange thickness of the HEM profile."""
    total_height: MM
    """Total height of the HEM profile."""
    web_thickness: MM
    """Web thickness of the HEM profile."""
    top_radius: MM
    """Top radius of the HEM profile."""
    bottom_radius: MM
    """Bottom radius of the HEM profile."""


HEM_PROFILES_DATABASE = {
    "HEM100": __HEMProfileParameters("HEM100", 106, 20, 106, 20, 120, 12, 12, 12),
    "HEM120": __HEMProfileParameters("HEM120", 126, 21, 126, 21, 140, 12.5, 12, 12),
    "HEM140": __HEMProfileParameters("HEM140", 146, 22, 146, 22, 160, 13, 12, 12),
    "HEM160": __HEMProfileParameters("HEM160", 166, 23, 166, 23, 180, 14, 15, 15),
    "HEM180": __HEMProfileParameters("HEM180", 186, 24, 186, 24, 200, 14.5, 15, 15),
    "HEM200": __HEMProfileParameters("HEM200", 206, 25, 206, 25, 220, 15, 18, 18),
    "HEM220": __HEMProfileParameters("HEM220", 226, 26, 226, 26, 240, 15.5, 18, 18),
    "HEM240": __HEMProfileParameters("HEM240", 248, 32, 248, 32, 270, 18, 21, 21),
    "HEM260": __HEMProfileParameters("HEM260", 268, 32.5, 268, 32.5, 290, 18, 24, 24),
    "HEM280": __HEMProfileParameters("HEM280", 288, 39, 288, 39, 310, 18.5, 24, 24),
    "HEM300": __HEMProfileParameters("HEM300", 310, 39, 310, 39, 340, 21, 27, 27),
    "HEM320": __HEMProfileParameters("HEM320", 309, 40, 309, 40, 359, 21, 27, 27),
    "HEM340": __HEMProfileParameters("HEM340", 309, 40, 309, 40, 377, 21, 27, 27),
    "HEM360": __HEMProfileParameters("HEM360", 308, 40, 308, 40, 395, 21, 27, 27),
    "HEM400": __HEMProfileParameters("HEM400", 307, 40, 307, 40, 432, 21, 27, 27),
    "HEM450": __HEMProfileParameters("HEM450", 307, 40, 307, 40, 478, 21, 27, 27),
    "HEM500": __HEMProfileParameters("HEM500", 306, 40, 306, 40, 524, 21, 27, 27),
    "HEM550": __HEMProfileParameters("HEM550", 306, 40, 306, 40, 572, 21, 27, 27),
    "HEM600": __HEMProfileParameters("HEM600", 305, 40, 305, 40, 620, 21, 27, 27),
    "HEM650": __HEMProfileParameters("HEM650", 305, 40, 305, 40, 668, 21, 27, 27),
    "HEM700": __HEMProfileParameters("HEM700", 304, 40, 304, 40, 716, 21, 27, 27),
    "HEM800": __HEMProfileParameters("HEM800", 303, 40, 303, 40, 814, 21, 30, 30),
    "HEM900": __HEMProfileParameters("HEM900", 302, 40, 302, 40, 910, 21, 30, 30),
    "HEM1000": __HEMProfileParameters("HEM1000", 302, 40, 302, 40, 1008, 21, 30, 30),
}


class HEM(metaclass=StandardProfileMeta):
    """Geometrical representation of HEM steel profiles.

    This class provides access to standard HEM profiles from a predefined database.
    Profiles can be accessed as class attributes using their standardized names.
    Each accessed profile returns a IProfile instance.

    Usage example
    -------------
        >>> profile = HEM.HEM200
    """

    _factory = IProfile
    """Factory class for creating standard HEM profiles."""
    _database = HEM_PROFILES_DATABASE
    """Database of standard HEM profile parameters."""



"""Tests for report_helpers.py helpers."""

import pytest

from blueprints.checks.check_result import CheckResult
from blueprints.utils.report import Report, report_helpers


@pytest.fixture
def fixture_report() -> Report:
    """Fixture for a real Report instance."""
    return Report()


def test_add_applied_documents(fixture_report: Report) -> None:
    """Test add_applied_documents helper."""
    docs = ["EN 1993-1-1", "EN 1990"]
    report_helpers.add_applied_documents(fixture_report, docs)
    content = fixture_report.content
    assert "Applied code documents" in content
    assert "The following documents were applied" in content
    for doc in docs:
        assert str(doc) in content


def test_add_applied_forces(fixture_report: Report) -> None:
    """Test add_applied_forces helper."""

    class DummyForces:
        n = 100.1234
        vy = 50.5678
        vz = 25.4321
        mx = 10.0
        my = 5.0
        mz = 1.0

    forces = DummyForces()
    report_helpers.add_applied_forces(fixture_report, forces, n=1)
    content = fixture_report.content
    assert "internal forces" in content.lower()
    assert "internal forces were applied" in content
    assert "Internal Force" in content
    assert "Normal force" in content
    assert "100.1" in content


def test_add_material_steel_info(fixture_report: Report) -> None:
    """Test add_material_steel_info helper."""

    class DummyMat:
        name = "S355"
        e_modulus = 210000

    class DummySteel:
        material = DummyMat()
        yield_strength = 355.5
        ultimate_strength = 510.2

    steel = DummySteel()
    report_helpers.add_material_steel_info(fixture_report, steel, n=0)
    content = fixture_report.content
    assert "material" in content.lower()
    assert "material properties" in content
    assert "Material" in content
    assert "S355" in content
    assert "Yield Strength" in content
    assert "356 MPa" in content
    assert "Elastic Modulus" in content
    assert "210000 MPa" in content


def test_add_section_properties_with_profile(fixture_report: Report) -> None:
    """Test add_section_properties with profile."""

    class DummySection:
        area = 123.456
        perimeter = 42.0

    class DummyProfile:
        name = "IPE 200"

    section = DummySection()
    profile = DummyProfile()
    report_helpers.add_section_properties(fixture_report, section, profile=profile, n=1, properties=["area", "perimeter", "not_existing"])
    content = fixture_report.content
    assert "section properties" in content
    assert "Profile" in content
    assert "IPE 200" in content
    assert "Area" in content
    assert "123.5" in content
    assert "Perimeter" in content
    assert "42.0" in content
    assert "Not_existing" in content
    assert "Not calculated" in content


def test_add_section_properties_without_profile(fixture_report: Report) -> None:
    """Test add_section_properties without profile."""

    class DummySection:
        area = 10.0

    section = DummySection()
    report_helpers.add_section_properties(fixture_report, section, profile=None, n=2, properties=["area"])
    content = fixture_report.content
    assert "Area" in content
    assert "10.00" in content


def test_add_unity_check_summary(fixture_report: Report) -> None:
    """Test add_unity_check_summary helper."""

    class DummyResult:
        def __init__(self, uc: float, ok: bool) -> None:
            self.unity_check = uc
            self.is_ok = ok

    class DummyCheck:
        def __init__(self, uc: float, ok: bool) -> None:
            self._res = DummyResult(uc, ok)

        def result(self) -> "DummyResult":
            return self._res

    checks = {
        "bending": DummyCheck(0.95, True),
        "shear": DummyCheck(1.05, False),
        "none": None,
        "check_result": CheckResult(is_ok=True),
    }
    report_helpers.add_unity_check_summary(report=fixture_report, calculations=checks, n=2)
    content = fixture_report.content
    assert "utilization" in content.lower()
    assert "Check" in content
    assert "Utilization" in content
    assert "Status" in content
    assert "Bending" in content
    assert "0.95" in content
    assert "OK" in content
    assert "Shear" in content
    assert "1.05" in content
    assert "NOT OK" in content
    assert "Overall result: NOT OK" in content


def test_add_unity_check_summary_empty(fixture_report: Report) -> None:
    """Test add_unity_check_summary with empty input."""
    report_helpers.add_unity_check_summary(fixture_report, {}, n=2)
    content = fixture_report.content
    assert "No checks were performed" in content
